"""Dependency injector singleton providers unit tests."""

import unittest

from dependency_injector import (
    providers,
    errors,
)

from .singleton_common import Example, _BaseSingletonTestCase


class SingletonTests(_BaseSingletonTestCase, unittest.TestCase):

    singleton_cls = providers.Singleton

    def test_repr(self):
        provider = self.singleton_cls(Example)

        self.assertEqual(repr(provider),
                         '<dependency_injector.providers.'
                         'Singleton({0}) at {1}>'.format(
                             repr(Example),
                             hex(id(provider))))


class DelegatedSingletonTests(_BaseSingletonTestCase, unittest.TestCase):

    singleton_cls = providers.DelegatedSingleton

    def test_is_delegated_provider(self):
        provider = self.singleton_cls(object)
        self.assertTrue(providers.is_delegated(provider))

    def test_repr(self):
        provider = self.singleton_cls(Example)

        self.assertEqual(repr(provider),
                         '<dependency_injector.providers.'
                         'DelegatedSingleton({0}) at {1}>'.format(
                             repr(Example),
                             hex(id(provider))))


class ThreadLocalSingletonTests(_BaseSingletonTestCase, unittest.TestCase):

    singleton_cls = providers.ThreadLocalSingleton

    def test_repr(self):
        provider = providers.ThreadLocalSingleton(Example)

        self.assertEqual(repr(provider),
                         '<dependency_injector.providers.'
                         'ThreadLocalSingleton({0}) at {1}>'.format(
                             repr(Example),
                             hex(id(provider))))

    def test_reset(self):
        provider = providers.ThreadLocalSingleton(Example)

        instance1 = provider()
        self.assertIsInstance(instance1, Example)

        provider.reset()

        instance2 = provider()
        self.assertIsInstance(instance2, Example)

        self.assertIsNot(instance1, instance2)

    def test_reset_clean(self):
        provider = providers.ThreadLocalSingleton(Example)
        instance1 = provider()

        provider.reset()
        provider.reset()

        instance2 = provider()
        self.assertIsNot(instance1, instance2)


class DelegatedThreadLocalSingletonTests(_BaseSingletonTestCase,
                                         unittest.TestCase):

    singleton_cls = providers.DelegatedThreadLocalSingleton

    def test_is_delegated_provider(self):
        provider = self.singleton_cls(object)
        self.assertTrue(providers.is_delegated(provider))

    def test_repr(self):
        provider = self.singleton_cls(Example)

        self.assertEqual(repr(provider),
                         '<dependency_injector.providers.'
                         'DelegatedThreadLocalSingleton({0}) at {1}>'.format(
                             repr(Example),
                             hex(id(provider))))


class ThreadSafeSingletonTests(_BaseSingletonTestCase, unittest.TestCase):

    singleton_cls = providers.ThreadSafeSingleton

    def test_repr(self):
        provider = self.singleton_cls(Example)

        self.assertEqual(repr(provider),
                         '<dependency_injector.providers.'
                         'ThreadSafeSingleton({0}) at {1}>'.format(
                             repr(Example),
                             hex(id(provider))))


class DelegatedThreadSafeSingletonTests(_BaseSingletonTestCase,
                                        unittest.TestCase):

    singleton_cls = providers.DelegatedThreadSafeSingleton

    def test_is_delegated_provider(self):
        provider = self.singleton_cls(object)
        self.assertTrue(providers.is_delegated(provider))

    def test_repr(self):
        provider = self.singleton_cls(Example)

        self.assertEqual(repr(provider),
                         '<dependency_injector.providers.'
                         'DelegatedThreadSafeSingleton({0}) at {1}>'.format(
                             repr(Example),
                             hex(id(provider))))


class AbstractSingletonTests(unittest.TestCase):

    def test_inheritance(self):
        self.assertIsInstance(providers.AbstractSingleton(Example),
                              providers.BaseSingleton)

    def test_call_overridden_by_singleton(self):
        provider = providers.AbstractSingleton(object)
        provider.override(providers.Singleton(Example))

        self.assertIsInstance(provider(), Example)

    def test_call_overridden_by_delegated_singleton(self):
        provider = providers.AbstractSingleton(object)
        provider.override(providers.DelegatedSingleton(Example))

        self.assertIsInstance(provider(), Example)

    def test_call_not_overridden(self):
        provider = providers.AbstractSingleton(object)

        with self.assertRaises(errors.Error):
            provider()

    def test_reset_overridden(self):
        provider = providers.AbstractSingleton(object)
        provider.override(providers.Singleton(Example))

        instance1 = provider()

        provider.reset()

        instance2 = provider()

        self.assertIsNot(instance1, instance2)
        self.assertIsInstance(instance1, Example)
        self.assertIsInstance(instance2, Example)

    def test_reset_not_overridden(self):
        provider = providers.AbstractSingleton(object)

        with self.assertRaises(errors.Error):
            provider.reset()

    def test_override_by_not_singleton(self):
        provider = providers.AbstractSingleton(object)

        with self.assertRaises(errors.Error):
            provider.override(providers.Factory(object))

    def test_repr(self):
        provider = providers.AbstractSingleton(Example)

        self.assertEqual(repr(provider),
                         '<dependency_injector.providers.'
                         'AbstractSingleton({0}) at {1}>'.format(
                             repr(Example),
                             hex(id(provider))))


class SingletonDelegateTests(unittest.TestCase):

    def setUp(self):
        self.delegated = providers.Singleton(Example)
        self.delegate = providers.SingletonDelegate(self.delegated)

    def test_is_delegate(self):
        self.assertIsInstance(self.delegate, providers.Delegate)

    def test_init_with_not_singleton(self):
        self.assertRaises(errors.Error,
                          providers.SingletonDelegate,
                          providers.Object(object()))



"""Dependency injector dynamic container unit tests for async resources."""
import asyncio

# Runtime import to get asyncutils module
import os
_TOP_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        '../',
    )),
)
import sys
sys.path.append(_TOP_DIR)

from asyncutils import AsyncTestCase

from dependency_injector import (
    containers,
    providers,
)


class AsyncResourcesTest(AsyncTestCase):

    def test_init_and_shutdown_ordering(self):
        """Test init and shutdown resources.

        Methods .init_resources() and .shutdown_resources() should respect resources dependencies.
        Initialization should first initialize resources without dependencies and then provide
        these resources to other resources. Resources shutdown should follow the same rule: first
        shutdown resources without initialized dependencies and then continue correspondingly
        until all resources are shutdown.
        """
        initialized_resources = []
        shutdown_resources = []

        async def _resource(name, delay, **_):
            await asyncio.sleep(delay)
            initialized_resources.append(name)

            yield name

            await asyncio.sleep(delay)
            shutdown_resources.append(name)

        class Container(containers.DeclarativeContainer):
            resource1 = providers.Resource(
                _resource,
                name='r1',
                delay=0.03,
            )
            resource2 = providers.Resource(
                _resource,
                name='r2',
                delay=0.02,
                r1=resource1,
            )
            resource3 = providers.Resource(
                _resource,
                name='r3',
                delay=0.01,
                r2=resource2,
            )

        container = Container()

        self._run(container.init_resources())
        self.assertEqual(initialized_resources, ['r1', 'r2', 'r3'])
        self.assertEqual(shutdown_resources, [])

        self._run(container.shutdown_resources())
        self.assertEqual(initialized_resources, ['r1', 'r2', 'r3'])
        self.assertEqual(shutdown_resources, ['r3', 'r2', 'r1'])

        self._run(container.init_resources())
        self.assertEqual(initialized_resources, ['r1', 'r2', 'r3', 'r1', 'r2', 'r3'])
        self.assertEqual(shutdown_resources, ['r3', 'r2', 'r1'])

        self._run(container.shutdown_resources())
        self.assertEqual(initialized_resources, ['r1', 'r2', 'r3', 'r1', 'r2', 'r3'])
        self.assertEqual(shutdown_resources, ['r3', 'r2', 'r1', 'r3', 'r2', 'r1'])

    def test_shutdown_circular_dependencies_breaker(self):
        async def _resource(name, **_):
            yield name

        class Container(containers.DeclarativeContainer):
            resource1 = providers.Resource(
                _resource,
                name='r1',
            )
            resource2 = providers.Resource(
                _resource,
                name='r2',
                r1=resource1,
            )
            resource3 = providers.Resource(
                _resource,
                name='r3',
                r2=resource2,
            )

        container = Container()
        self._run(container.init_resources())

        # Create circular dependency after initialization (r3 -> r2 -> r1 -> r3 -> ...)
        container.resource1.add_kwargs(r3=container.resource3)

        with self.assertRaises(RuntimeError) as context:
            self._run(container.shutdown_resources())
        self.assertEqual(str(context.exception), 'Unable to resolve resources shutdown order')

    def test_shutdown_sync_and_async_ordering(self):
        initialized_resources = []
        shutdown_resources = []

        def _sync_resource(name, **_):
            initialized_resources.append(name)
            yield name
            shutdown_resources.append(name)

        async def _async_resource(name, **_):
            initialized_resources.append(name)
            yield name
            shutdown_resources.append(name)

        class Container(containers.DeclarativeContainer):
            resource1 = providers.Resource(
                _sync_resource,
                name='r1',
            )
            resource2 = providers.Resource(
                _sync_resource,
                name='r2',
                r1=resource1,
            )
            resource3 = providers.Resource(
                _async_resource,
                name='r3',
                r2=resource2,
            )

        container = Container()

        self._run(container.init_resources())
        self.assertEqual(initialized_resources, ['r1', 'r2', 'r3'])
        self.assertEqual(shutdown_resources, [])

        self._run(container.shutdown_resources())
        self.assertEqual(initialized_resources, ['r1', 'r2', 'r3'])
        self.assertEqual(shutdown_resources, ['r3', 'r2', 'r1'])

        self._run(container.init_resources())
        self.assertEqual(initialized_resources, ['r1', 'r2', 'r3', 'r1', 'r2', 'r3'])
        self.assertEqual(shutdown_resources, ['r3', 'r2', 'r1'])

        self._run(container.shutdown_resources())
        self.assertEqual(initialized_resources, ['r1', 'r2', 'r3', 'r1', 'r2', 'r3'])
        self.assertEqual(shutdown_resources, ['r3', 'r2', 'r1', 'r3', 'r2', 'r1'])



import contextlib
from decimal import Decimal
import importlib
import unittest

from dependency_injector.wiring import (
    wire,
    Provide,
    Provider,
    Closing,
    register_loader_containers,
    unregister_loader_containers,
)
from dependency_injector import containers, providers, errors

# Runtime import to avoid syntax errors in samples on Python < 3.5
import os
_TOP_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        '../',
    )),
)
_SAMPLES_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        '../samples/',
    )),
)
import sys
sys.path.append(_TOP_DIR)
sys.path.append(_SAMPLES_DIR)

from asyncutils import AsyncTestCase

from wiringstringidssamples import module, package
from wiringstringidssamples.service import Service
from wiringstringidssamples.container import Container, SubContainer


class WiringTest(unittest.TestCase):

    container: Container

    def setUp(self) -> None:
        self.container = Container(config={'a': {'b': {'c': 10}}})
        self.container.wire(
            modules=[module],
            packages=[package],
        )
        self.addCleanup(self.container.unwire)

    def test_package_lookup(self):
        from wiringstringidssamples.package import test_package_function
        service = test_package_function()
        self.assertIsInstance(service, Service)

    def test_package_subpackage_lookup(self):
        from wiringstringidssamples.package.subpackage import test_package_function
        service = test_package_function()
        self.assertIsInstance(service, Service)

    def test_package_submodule_lookup(self):
        from wiringstringidssamples.package.subpackage.submodule import test_function
        service = test_function()
        self.assertIsInstance(service, Service)

    def test_module_attributes_wiring(self):
        self.assertIsInstance(module.service, Service)
        self.assertIsInstance(module.service_provider(), Service)
        self.assertIsInstance(module.undefined, Provide)

    def test_class_wiring(self):
        test_class_object = module.TestClass()
        self.assertIsInstance(test_class_object.service, Service)

    def test_class_wiring_context_arg(self):
        test_service = self.container.service()

        test_class_object = module.TestClass(service=test_service)
        self.assertIs(test_class_object.service, test_service)

    def test_class_method_wiring(self):
        test_class_object = module.TestClass()
        service = test_class_object.method()
        self.assertIsInstance(service, Service)

    def test_class_classmethod_wiring(self):
        service = module.TestClass.class_method()
        self.assertIsInstance(service, Service)

    def test_instance_classmethod_wiring(self):
        instance = module.TestClass()
        service = instance.class_method()
        self.assertIsInstance(service, Service)

    def test_class_staticmethod_wiring(self):
        service = module.TestClass.static_method()
        self.assertIsInstance(service, Service)

    def test_instance_staticmethod_wiring(self):
        instance = module.TestClass()
        service = instance.static_method()
        self.assertIsInstance(service, Service)

    def test_class_attribute_wiring(self):
        self.assertIsInstance(module.TestClass.service, Service)
        self.assertIsInstance(module.TestClass.service_provider(), Service)
        self.assertIsInstance(module.TestClass.undefined, Provide)

    def test_function_wiring(self):
        service = module.test_function()
        self.assertIsInstance(service, Service)

    def test_function_wiring_context_arg(self):
        test_service = self.container.service()

        service = module.test_function(service=test_service)
        self.assertIs(service, test_service)

    def test_function_wiring_provider(self):
        service = module.test_function_provider()
        self.assertIsInstance(service, Service)

    def test_function_wiring_provider_context_arg(self):
        test_service = self.container.service()

        service = module.test_function_provider(service_provider=lambda: test_service)
        self.assertIs(service, test_service)

    def test_configuration_option(self):
        (
            value_int,
            value_float,
            value_str,
            value_decimal,
            value_required,
            value_required_int,
            value_required_float,
            value_required_str,
            value_required_decimal,
        ) = module.test_config_value()

        self.assertEqual(value_int, 10)
        self.assertEqual(value_float, 10.0)
        self.assertEqual(value_str, '10')
        self.assertEqual(value_decimal, Decimal(10))
        self.assertEqual(value_required, 10)
        self.assertEqual(value_required_int, 10)
        self.assertEqual(value_required_float, 10.0)
        self.assertEqual(value_required_str, '10')
        self.assertEqual(value_required_decimal, Decimal(10))

    def test_configuration_option_required_undefined(self):
        self.container.config.reset_override()
        with self.assertRaisesRegex(errors.Error, 'Undefined configuration option "config.a.b.c"'):
            module.test_config_value_required_undefined()

    def test_provide_provider(self):
        service = module.test_provide_provider()
        self.assertIsInstance(service, Service)

    def test_provided_instance(self):
        class TestService:
            foo = {
                'bar': lambda: 10,
            }

        with self.container.service.override(TestService()):
            some_value = module.test_provided_instance()
        self.assertEqual(some_value, 10)

    def test_subcontainer(self):
        some_value = module.test_subcontainer_provider()
        self.assertEqual(some_value, 1)

    def test_config_invariant(self):
        config = {
            'option': {
                'a': 1,
                'b': 2,
            },
            'switch': 'a',
        }
        self.container.config.from_dict(config)

        value_default = module.test_config_invariant()
        self.assertEqual(value_default, 1)

        with self.container.config.switch.override('a'):
            value_a = module.test_config_invariant()
        self.assertEqual(value_a, 1)

        with self.container.config.switch.override('b'):
            value_b = module.test_config_invariant()
        self.assertEqual(value_b, 2)

    def test_wire_with_class_error(self):
        with self.assertRaises(Exception):
            wire(
                container=Container,
                modules=[module],
            )

    def test_unwire_function(self):
        self.container.unwire()
        self.assertIsInstance(module.test_function(), Provide)

    def test_unwire_class(self):
        self.container.unwire()
        test_class_object = module.TestClass()
        self.assertIsInstance(test_class_object.service, Provide)

    def test_unwire_class_method(self):
        self.container.unwire()
        test_class_object = module.TestClass()
        self.assertIsInstance(test_class_object.method(), Provide)

    def test_unwire_package_function(self):
        self.container.unwire()
        from wiringstringidssamples.package.subpackage.submodule import test_function
        self.assertIsInstance(test_function(), Provide)

    def test_unwire_package_function_by_reference(self):
        from wiringstringidssamples.package.subpackage import submodule
        self.container.unwire()
        self.assertIsInstance(submodule.test_function(), Provide)

    def test_unwire_module_attributes(self):
        self.container.unwire()
        self.assertIsInstance(module.service, Provide)
        self.assertIsInstance(module.service_provider, Provider)
        self.assertIsInstance(module.undefined, Provide)

    def test_unwire_class_attributes(self):
        self.container.unwire()
        self.assertIsInstance(module.TestClass.service, Provide)
        self.assertIsInstance(module.TestClass.service_provider, Provider)
        self.assertIsInstance(module.TestClass.undefined, Provide)

    def test_wire_multiple_containers(self):
        sub_container = SubContainer()
        sub_container.wire(
            modules=[module],
            packages=[package],
        )
        self.addCleanup(sub_container.unwire)

        service, some_value = module.test_provide_from_different_containers()

        self.assertIsInstance(service, Service)
        self.assertEqual(some_value, 1)

    def test_closing_resource(self):
        from wiringstringidssamples import resourceclosing

        resourceclosing.Service.reset_counter()

        container = resourceclosing.Container()
        container.wire(modules=[resourceclosing])
        self.addCleanup(container.unwire)

        result_1 = resourceclosing.test_function()
        self.assertIsInstance(result_1, resourceclosing.Service)
        self.assertEqual(result_1.init_counter, 1)
        self.assertEqual(result_1.shutdown_counter, 1)

        result_2 = resourceclosing.test_function()
        self.assertIsInstance(result_2, resourceclosing.Service)
        self.assertEqual(result_2.init_counter, 2)
        self.assertEqual(result_2.shutdown_counter, 2)

        self.assertIsNot(result_1, result_2)

    def test_closing_resource_context(self):
        from wiringstringidssamples import resourceclosing

        resourceclosing.Service.reset_counter()
        service = resourceclosing.Service()

        container = resourceclosing.Container()
        container.wire(modules=[resourceclosing])
        self.addCleanup(container.unwire)

        result_1 = resourceclosing.test_function(service=service)
        self.assertIs(result_1, service)
        self.assertEqual(result_1.init_counter, 0)
        self.assertEqual(result_1.shutdown_counter, 0)

        result_2 = resourceclosing.test_function(service=service)
        self.assertIs(result_2, service)
        self.assertEqual(result_2.init_counter, 0)
        self.assertEqual(result_2.shutdown_counter, 0)

    def test_class_decorator(self):
        service = module.test_class_decorator()
        self.assertIsInstance(service, Service)

    def test_container(self):
        service = module.test_container()
        self.assertIsInstance(service, Service)


class WiringAndFastAPITest(unittest.TestCase):

    container: Container

    def test_bypass_marker_injection(self):
        container = Container()
        container.wire(modules=[module])
        self.addCleanup(container.unwire)

        service = module.test_function(service=Provide[Container.service])
        self.assertIsInstance(service, Service)

    def test_closing_resource_bypass_marker_injection(self):
        from wiringstringidssamples import resourceclosing

        resourceclosing.Service.reset_counter()

        container = resourceclosing.Container()
        container.wire(modules=[resourceclosing])
        self.addCleanup(container.unwire)

        result_1 = resourceclosing.test_function(
            service=Closing[Provide[resourceclosing.Container.service]],
        )
        self.assertIsInstance(result_1, resourceclosing.Service)
        self.assertEqual(result_1.init_counter, 1)
        self.assertEqual(result_1.shutdown_counter, 1)

        result_2 = resourceclosing.test_function(
            service=Closing[Provide[resourceclosing.Container.service]],
        )
        self.assertIsInstance(result_2, resourceclosing.Service)
        self.assertEqual(result_2.init_counter, 2)
        self.assertEqual(result_2.shutdown_counter, 2)

        self.assertIsNot(result_1, result_2)


class WireDynamicContainerTest(unittest.TestCase):

    def test_wire(self):
        sub = containers.DynamicContainer()
        sub.int_object = providers.Object(1)

        container = containers.DynamicContainer()
        container.config = providers.Configuration()
        container.service = providers.Factory(Service)
        container.sub = sub

        container.wire(
            modules=[module],
            packages=[package],
        )
        self.addCleanup(container.unwire)

        service = module.test_function()
        self.assertIsInstance(service, Service)


class WiringAsyncInjectionsTest(AsyncTestCase):

    def test_async_injections(self):
        from wiringstringidssamples import asyncinjections

        container = asyncinjections.Container()
        container.wire(modules=[asyncinjections])
        self.addCleanup(container.unwire)

        asyncinjections.resource1.reset_counters()
        asyncinjections.resource2.reset_counters()

        resource1, resource2 = self._run(asyncinjections.async_injection())

        self.assertIs(resource1, asyncinjections.resource1)
        self.assertEqual(asyncinjections.resource1.init_counter, 1)
        self.assertEqual(asyncinjections.resource1.shutdown_counter, 0)

        self.assertIs(resource2, asyncinjections.resource2)
        self.assertEqual(asyncinjections.resource2.init_counter, 1)
        self.assertEqual(asyncinjections.resource2.shutdown_counter, 0)

    def test_async_injections_with_closing(self):
        from wiringstringidssamples import asyncinjections

        container = asyncinjections.Container()
        container.wire(modules=[asyncinjections])
        self.addCleanup(container.unwire)

        asyncinjections.resource1.reset_counters()
        asyncinjections.resource2.reset_counters()

        resource1, resource2 = self._run(asyncinjections.async_injection_with_closing())

        self.assertIs(resource1, asyncinjections.resource1)
        self.assertEqual(asyncinjections.resource1.init_counter, 1)
        self.assertEqual(asyncinjections.resource1.shutdown_counter, 1)

        self.assertIs(resource2, asyncinjections.resource2)
        self.assertEqual(asyncinjections.resource2.init_counter, 1)
        self.assertEqual(asyncinjections.resource2.shutdown_counter, 1)

        resource1, resource2 = self._run(asyncinjections.async_injection_with_closing())

        self.assertIs(resource1, asyncinjections.resource1)
        self.assertEqual(asyncinjections.resource1.init_counter, 2)
        self.assertEqual(asyncinjections.resource1.shutdown_counter, 2)

        self.assertIs(resource2, asyncinjections.resource2)
        self.assertEqual(asyncinjections.resource2.init_counter, 2)
        self.assertEqual(asyncinjections.resource2.shutdown_counter, 2)


class AutoLoaderTest(unittest.TestCase):

    container: Container

    def setUp(self) -> None:
        self.container = Container(config={'a': {'b': {'c': 10}}})
        importlib.reload(module)

    def tearDown(self) -> None:
        with contextlib.suppress(ValueError):
            unregister_loader_containers(self.container)

        self.container.unwire()

    @classmethod
    def tearDownClass(cls) -> None:
        importlib.reload(module)

    def test_register_container(self):
        register_loader_containers(self.container)
        importlib.reload(module)

        service = module.test_function()
        self.assertIsInstance(service, Service)



import sys

from dependency_injector import providers, errors


class Example(object):

    def __init__(self, init_arg1=None, init_arg2=None, init_arg3=None,
                 init_arg4=None):
        self.init_arg1 = init_arg1
        self.init_arg2 = init_arg2
        self.init_arg3 = init_arg3
        self.init_arg4 = init_arg4

        self.attribute1 = None
        self.attribute2 = None


class _BaseSingletonTestCase(object):

    singleton_cls = None

    def test_is_provider(self):
        self.assertTrue(providers.is_provider(self.singleton_cls(Example)))

    def test_init_with_callable(self):
        self.assertTrue(self.singleton_cls(credits))

    def test_init_with_not_callable(self):
        self.assertRaises(errors.Error, self.singleton_cls, 123)

    def test_init_optional_provides(self):
        provider = self.singleton_cls()
        provider.set_provides(object)
        self.assertIs(provider.provides, object)
        self.assertIsInstance(provider(), object)

    def test_set_provides_returns_self(self):
        provider = self.singleton_cls()
        self.assertIs(provider.set_provides(object), provider)

    def test_init_with_valid_provided_type(self):
        class ExampleProvider(self.singleton_cls):
            provided_type = Example

        example_provider = ExampleProvider(Example, 1, 2)

        self.assertIsInstance(example_provider(), Example)

    def test_init_with_valid_provided_subtype(self):
        class ExampleProvider(self.singleton_cls):
            provided_type = Example

        class NewExampe(Example):
            pass

        example_provider = ExampleProvider(NewExampe, 1, 2)

        self.assertIsInstance(example_provider(), NewExampe)

    def test_init_with_invalid_provided_type(self):
        class ExampleProvider(self.singleton_cls):
            provided_type = Example

        with self.assertRaises(errors.Error):
            ExampleProvider(list)

    def test_provided_instance_provider(self):
        provider = providers.Singleton(Example)
        self.assertIsInstance(provider.provided, providers.ProvidedInstance)

    def test_call(self):
        provider = self.singleton_cls(Example)

        instance1 = provider()
        instance2 = provider()

        self.assertIs(instance1, instance2)
        self.assertIsInstance(instance1, Example)
        self.assertIsInstance(instance2, Example)

    def test_call_with_init_positional_args(self):
        provider = self.singleton_cls(Example, 'i1', 'i2')

        instance1 = provider()
        instance2 = provider()

        self.assertEqual(instance1.init_arg1, 'i1')
        self.assertEqual(instance1.init_arg2, 'i2')

        self.assertEqual(instance2.init_arg1, 'i1')
        self.assertEqual(instance2.init_arg2, 'i2')

        self.assertIs(instance1, instance2)
        self.assertIsInstance(instance1, Example)
        self.assertIsInstance(instance2, Example)

    def test_call_with_init_keyword_args(self):
        provider = self.singleton_cls(Example, init_arg1='i1', init_arg2='i2')

        instance1 = provider()
        instance2 = provider()

        self.assertEqual(instance1.init_arg1, 'i1')
        self.assertEqual(instance1.init_arg2, 'i2')

        self.assertEqual(instance2.init_arg1, 'i1')
        self.assertEqual(instance2.init_arg2, 'i2')

        self.assertIs(instance1, instance2)
        self.assertIsInstance(instance1, Example)
        self.assertIsInstance(instance2, Example)

    def test_call_with_init_positional_and_keyword_args(self):
        provider = self.singleton_cls(Example, 'i1', init_arg2='i2')

        instance1 = provider()
        instance2 = provider()

        self.assertEqual(instance1.init_arg1, 'i1')
        self.assertEqual(instance1.init_arg2, 'i2')

        self.assertEqual(instance2.init_arg1, 'i1')
        self.assertEqual(instance2.init_arg2, 'i2')

        self.assertIs(instance1, instance2)
        self.assertIsInstance(instance1, Example)
        self.assertIsInstance(instance2, Example)

    def test_call_with_attributes(self):
        provider = self.singleton_cls(Example)
        provider.add_attributes(attribute1='a1', attribute2='a2')

        instance1 = provider()
        instance2 = provider()

        self.assertEqual(instance1.attribute1, 'a1')
        self.assertEqual(instance1.attribute2, 'a2')

        self.assertEqual(instance2.attribute1, 'a1')
        self.assertEqual(instance2.attribute2, 'a2')

        self.assertIs(instance1, instance2)
        self.assertIsInstance(instance1, Example)
        self.assertIsInstance(instance2, Example)

    def test_call_with_context_args(self):
        provider = self.singleton_cls(Example)

        instance = provider(11, 22)

        self.assertEqual(instance.init_arg1, 11)
        self.assertEqual(instance.init_arg2, 22)

    def test_call_with_context_kwargs(self):
        provider = self.singleton_cls(Example, init_arg1=1)

        instance1 = provider(init_arg2=22)
        self.assertEqual(instance1.init_arg1, 1)
        self.assertEqual(instance1.init_arg2, 22)

        # Instance is created earlier
        instance1 = provider(init_arg1=11, init_arg2=22)
        self.assertEqual(instance1.init_arg1, 1)
        self.assertEqual(instance1.init_arg2, 22)

    def test_call_with_context_args_and_kwargs(self):
        provider = self.singleton_cls(Example, 11)

        instance = provider(22, init_arg3=33, init_arg4=44)

        self.assertEqual(instance.init_arg1, 11)
        self.assertEqual(instance.init_arg2, 22)
        self.assertEqual(instance.init_arg3, 33)
        self.assertEqual(instance.init_arg4, 44)

    def test_fluent_interface(self):
        provider = self.singleton_cls(Example) \
            .add_args(1, 2) \
            .add_kwargs(init_arg3=3, init_arg4=4) \
            .add_attributes(attribute1=5, attribute2=6)

        instance = provider()

        self.assertEqual(instance.init_arg1, 1)
        self.assertEqual(instance.init_arg2, 2)
        self.assertEqual(instance.init_arg3, 3)
        self.assertEqual(instance.init_arg4, 4)
        self.assertEqual(instance.attribute1, 5)
        self.assertEqual(instance.attribute2, 6)

    def test_set_args(self):
        provider = self.singleton_cls(Example) \
            .add_args(1, 2) \
            .set_args(3, 4)
        self.assertEqual(provider.args, (3, 4))

    def test_set_kwargs(self):
        provider = self.singleton_cls(Example) \
            .add_kwargs(init_arg3=3, init_arg4=4) \
            .set_kwargs(init_arg3=4, init_arg4=5)
        self.assertEqual(provider.kwargs, dict(init_arg3=4, init_arg4=5))

    def test_set_attributes(self):
        provider = self.singleton_cls(Example) \
            .add_attributes(attribute1=5, attribute2=6) \
            .set_attributes(attribute1=6, attribute2=7)
        self.assertEqual(provider.attributes, dict(attribute1=6, attribute2=7))

    def test_clear_args(self):
        provider = self.singleton_cls(Example) \
            .add_args(1, 2) \
            .clear_args()
        self.assertEqual(provider.args, tuple())

    def test_clear_kwargs(self):
        provider = self.singleton_cls(Example) \
            .add_kwargs(init_arg3=3, init_arg4=4) \
            .clear_kwargs()
        self.assertEqual(provider.kwargs, dict())

    def test_clear_attributes(self):
        provider = self.singleton_cls(Example) \
            .add_attributes(attribute1=5, attribute2=6) \
            .clear_attributes()
        self.assertEqual(provider.attributes, dict())

    def test_call_overridden(self):
        provider = self.singleton_cls(Example)
        overriding_provider1 = self.singleton_cls(dict)
        overriding_provider2 = self.singleton_cls(list)

        provider.override(overriding_provider1)
        provider.override(overriding_provider2)

        instance1 = provider()
        instance2 = provider()

        self.assertIs(instance1, instance2)
        self.assertIsInstance(instance1, list)
        self.assertIsInstance(instance2, list)

    def test_deepcopy(self):
        provider = self.singleton_cls(Example)

        provider_copy = providers.deepcopy(provider)

        self.assertIsNot(provider, provider_copy)
        self.assertIs(provider.cls, provider_copy.cls)
        self.assertIsInstance(provider, self.singleton_cls)

    def test_deepcopy_from_memo(self):
        provider = self.singleton_cls(Example)
        provider_copy_memo = self.singleton_cls(Example)

        provider_copy = providers.deepcopy(
            provider, memo={id(provider): provider_copy_memo})

        self.assertIs(provider_copy, provider_copy_memo)

    def test_deepcopy_args(self):
        provider = self.singleton_cls(Example)
        dependent_provider1 = self.singleton_cls(list)
        dependent_provider2 = self.singleton_cls(dict)

        provider.add_args(dependent_provider1, dependent_provider2)

        provider_copy = providers.deepcopy(provider)
        dependent_provider_copy1 = provider_copy.args[0]
        dependent_provider_copy2 = provider_copy.args[1]

        self.assertNotEqual(provider.args, provider_copy.args)

        self.assertIs(dependent_provider1.cls, dependent_provider_copy1.cls)
        self.assertIsNot(dependent_provider1, dependent_provider_copy1)

        self.assertIs(dependent_provider2.cls, dependent_provider_copy2.cls)
        self.assertIsNot(dependent_provider2, dependent_provider_copy2)

    def test_deepcopy_kwargs(self):
        provider = self.singleton_cls(Example)
        dependent_provider1 = self.singleton_cls(list)
        dependent_provider2 = self.singleton_cls(dict)

        provider.add_kwargs(a1=dependent_provider1, a2=dependent_provider2)

        provider_copy = providers.deepcopy(provider)
        dependent_provider_copy1 = provider_copy.kwargs['a1']
        dependent_provider_copy2 = provider_copy.kwargs['a2']

        self.assertNotEqual(provider.kwargs, provider_copy.kwargs)

        self.assertIs(dependent_provider1.cls, dependent_provider_copy1.cls)
        self.assertIsNot(dependent_provider1, dependent_provider_copy1)

        self.assertIs(dependent_provider2.cls, dependent_provider_copy2.cls)
        self.assertIsNot(dependent_provider2, dependent_provider_copy2)

    def test_deepcopy_attributes(self):
        provider = self.singleton_cls(Example)
        dependent_provider1 = self.singleton_cls(list)
        dependent_provider2 = self.singleton_cls(dict)

        provider.add_attributes(a1=dependent_provider1, a2=dependent_provider2)

        provider_copy = providers.deepcopy(provider)
        dependent_provider_copy1 = provider_copy.attributes['a1']
        dependent_provider_copy2 = provider_copy.attributes['a2']

        self.assertNotEqual(provider.attributes, provider_copy.attributes)

        self.assertIs(dependent_provider1.cls, dependent_provider_copy1.cls)
        self.assertIsNot(dependent_provider1, dependent_provider_copy1)

        self.assertIs(dependent_provider2.cls, dependent_provider_copy2.cls)
        self.assertIsNot(dependent_provider2, dependent_provider_copy2)

    def test_deepcopy_overridden(self):
        provider = self.singleton_cls(Example)
        object_provider = providers.Object(object())

        provider.override(object_provider)

        provider_copy = providers.deepcopy(provider)
        object_provider_copy = provider_copy.overridden[0]

        self.assertIsNot(provider, provider_copy)
        self.assertIs(provider.cls, provider_copy.cls)
        self.assertIsInstance(provider, self.singleton_cls)

        self.assertIsNot(object_provider, object_provider_copy)
        self.assertIsInstance(object_provider_copy, providers.Object)

    def test_deepcopy_with_sys_streams(self):
        provider = providers.Singleton(Example)
        provider.add_args(sys.stdin)
        provider.add_kwargs(a2=sys.stdout)
        provider.add_attributes(a3=sys.stderr)

        provider_copy = providers.deepcopy(provider)

        self.assertIsNot(provider, provider_copy)
        self.assertIsInstance(provider_copy, providers.Singleton)
        self.assertIs(provider.args[0], sys.stdin)
        self.assertIs(provider.kwargs['a2'], sys.stdout)
        self.assertIs(provider.attributes['a3'], sys.stderr)

    def test_reset(self):
        provider = self.singleton_cls(object)

        instance1 = provider()
        self.assertIsInstance(instance1, object)

        provider.reset()

        instance2 = provider()
        self.assertIsInstance(instance2, object)

        self.assertIsNot(instance1, instance2)

    def test_reset_with_singleton(self):
        dependent_singleton = providers.Singleton(object)
        provider = self.singleton_cls(dict, dependency=dependent_singleton)

        dependent_instance = dependent_singleton()
        instance1 = provider()
        self.assertIs(instance1['dependency'], dependent_instance)

        provider.reset()

        instance2 = provider()
        self.assertIs(instance1['dependency'], dependent_instance)

        self.assertIsNot(instance1, instance2)

    def test_reset_context_manager(self):
        singleton = self.singleton_cls(object)

        instance1 = singleton()
        with singleton.reset():
            instance2 = singleton()
        instance3 = singleton()
        self.assertEqual(len({instance1, instance2, instance3}), 3)

    def test_reset_context_manager_as_attribute(self):
        singleton = self.singleton_cls(object)

        with singleton.reset() as alias:
            pass

        self.assertIs(singleton, alias)

    def test_full_reset(self):
        dependent_singleton = providers.Singleton(object)
        provider = self.singleton_cls(dict, dependency=dependent_singleton)

        dependent_instance1 = dependent_singleton()
        instance1 = provider()
        self.assertIs(instance1['dependency'], dependent_instance1)

        provider.full_reset()

        dependent_instance2 = dependent_singleton()
        instance2 = provider()
        self.assertIsNot(instance2['dependency'], dependent_instance1)
        self.assertIsNot(dependent_instance1, dependent_instance2)
        self.assertIsNot(instance1, instance2)

    def test_full_reset_context_manager(self):
        class Item:
            def __init__(self, dependency):
                self.dependency = dependency

        dependent_singleton = providers.Singleton(object)
        singleton = self.singleton_cls(Item, dependency=dependent_singleton)

        instance1 = singleton()
        with singleton.full_reset():
            instance2 = singleton()
        instance3 = singleton()

        self.assertEqual(len({instance1, instance2, instance3}), 3)
        self.assertEqual(
            len({instance1.dependency, instance2.dependency, instance3.dependency}),
            3,
        )

    def test_full_reset_context_manager_as_attribute(self):
        singleton = self.singleton_cls(object)

        with singleton.full_reset() as alias:
            pass

        self.assertIs(singleton, alias)



"""Dependency injector dict provider unit tests."""

import sys

import unittest

from dependency_injector import providers


class DictTests(unittest.TestCase):

    def test_is_provider(self):
        self.assertTrue(providers.is_provider(providers.Dict()))

    def test_provided_instance_provider(self):
        provider = providers.Dict()
        self.assertIsInstance(provider.provided, providers.ProvidedInstance)

    def test_init_with_non_string_keys(self):
        a1 = object()
        a2 = object()
        provider = providers.Dict({a1: 'i1', a2: 'i2'})

        dict1 = provider()
        dict2 = provider()

        self.assertEqual(dict1, {a1: 'i1', a2: 'i2'})
        self.assertEqual(dict2, {a1: 'i1', a2: 'i2'})

        self.assertIsNot(dict1, dict2)

    def test_init_with_string_and_non_string_keys(self):
        a1 = object()
        provider = providers.Dict({a1: 'i1'}, a2='i2')

        dict1 = provider()
        dict2 = provider()

        self.assertEqual(dict1, {a1: 'i1', 'a2': 'i2'})
        self.assertEqual(dict2, {a1: 'i1', 'a2': 'i2'})

        self.assertIsNot(dict1, dict2)

    def test_call_with_init_keyword_args(self):
        provider = providers.Dict(a1='i1', a2='i2')

        dict1 = provider()
        dict2 = provider()

        self.assertEqual(dict1, {'a1': 'i1', 'a2': 'i2'})
        self.assertEqual(dict2, {'a1': 'i1', 'a2': 'i2'})

        self.assertIsNot(dict1, dict2)

    def test_call_with_context_keyword_args(self):
        provider = providers.Dict(a1='i1', a2='i2')
        self.assertEqual(
            provider(a3='i3', a4='i4'),
            {'a1': 'i1', 'a2': 'i2', 'a3': 'i3', 'a4': 'i4'},
        )

    def test_call_with_provider(self):
        provider = providers.Dict(
            a1=providers.Factory(str, 'i1'),
            a2=providers.Factory(str, 'i2'),
        )
        self.assertEqual(provider(), {'a1': 'i1', 'a2': 'i2'})

    def test_fluent_interface(self):
        provider = providers.Dict() \
            .add_kwargs(a1='i1', a2='i2')
        self.assertEqual(provider(), {'a1': 'i1', 'a2': 'i2'})

    def test_add_kwargs(self):
        provider = providers.Dict() \
            .add_kwargs(a1='i1') \
            .add_kwargs(a2='i2')
        self.assertEqual(provider.kwargs, {'a1': 'i1', 'a2': 'i2'})

    def test_add_kwargs_non_string_keys(self):
        a1 = object()
        a2 = object()
        provider = providers.Dict() \
            .add_kwargs({a1: 'i1'}) \
            .add_kwargs({a2: 'i2'})
        self.assertEqual(provider.kwargs, {a1: 'i1', a2: 'i2'})

    def test_add_kwargs_string_and_non_string_keys(self):
        a2 = object()
        provider = providers.Dict() \
            .add_kwargs(a1='i1') \
            .add_kwargs({a2: 'i2'})
        self.assertEqual(provider.kwargs, {'a1': 'i1', a2: 'i2'})

    def test_set_kwargs(self):
        provider = providers.Dict() \
            .add_kwargs(a1='i1', a2='i2') \
            .set_kwargs(a3='i3', a4='i4')
        self.assertEqual(provider.kwargs, {'a3': 'i3', 'a4': 'i4'})

    def test_set_kwargs_non_string_keys(self):
        a3 = object()
        a4 = object()
        provider = providers.Dict() \
            .add_kwargs(a1='i1', a2='i2') \
            .set_kwargs({a3: 'i3', a4: 'i4'})
        self.assertEqual(provider.kwargs, {a3: 'i3', a4: 'i4'})

    def test_set_kwargs_string_and_non_string_keys(self):
        a3 = object()
        provider = providers.Dict() \
            .add_kwargs(a1='i1', a2='i2') \
            .set_kwargs({a3: 'i3'}, a4='i4')
        self.assertEqual(provider.kwargs, {a3: 'i3', 'a4': 'i4'})

    def test_clear_kwargs(self):
        provider = providers.Dict() \
            .add_kwargs(a1='i1', a2='i2') \
            .clear_kwargs()
        self.assertEqual(provider.kwargs, {})

    def test_call_overridden(self):
        provider = providers.Dict(a1='i1', a2='i2')
        overriding_provider1 = providers.Dict(a2='i2', a3='i3')
        overriding_provider2 = providers.Dict(a3='i3', a4='i4')

        provider.override(overriding_provider1)
        provider.override(overriding_provider2)

        instance1 = provider()
        instance2 = provider()

        self.assertIsNot(instance1, instance2)
        self.assertEqual(instance1, {'a3': 'i3', 'a4': 'i4'})
        self.assertEqual(instance2, {'a3': 'i3', 'a4': 'i4'})

    def test_deepcopy(self):
        provider = providers.Dict(a1='i1', a2='i2')

        provider_copy = providers.deepcopy(provider)

        self.assertIsNot(provider, provider_copy)
        self.assertEqual(provider.kwargs, provider_copy.kwargs)
        self.assertIsInstance(provider, providers.Dict)

    def test_deepcopy_from_memo(self):
        provider = providers.Dict(a1='i1', a2='i2')
        provider_copy_memo = providers.Dict(a1='i1', a2='i2')

        provider_copy = providers.deepcopy(
            provider,
            memo={id(provider): provider_copy_memo},
        )

        self.assertIs(provider_copy, provider_copy_memo)

    def test_deepcopy_kwargs(self):
        provider = providers.Dict()
        dependent_provider1 = providers.Factory(list)
        dependent_provider2 = providers.Factory(dict)

        provider.add_kwargs(d1=dependent_provider1, d2=dependent_provider2)

        provider_copy = providers.deepcopy(provider)
        dependent_provider_copy1 = provider_copy.kwargs['d1']
        dependent_provider_copy2 = provider_copy.kwargs['d2']

        self.assertNotEqual(provider.kwargs, provider_copy.kwargs)

        self.assertIs(dependent_provider1.cls, dependent_provider_copy1.cls)
        self.assertIsNot(dependent_provider1, dependent_provider_copy1)

        self.assertIs(dependent_provider2.cls, dependent_provider_copy2.cls)
        self.assertIsNot(dependent_provider2, dependent_provider_copy2)

    def test_deepcopy_kwargs_non_string_keys(self):
        a1 = object()
        a2 = object()

        dependent_provider1 = providers.Factory(list)
        dependent_provider2 = providers.Factory(dict)

        provider = providers.Dict({a1: dependent_provider1, a2: dependent_provider2})

        provider_copy = providers.deepcopy(provider)
        dependent_provider_copy1 = provider_copy.kwargs[a1]
        dependent_provider_copy2 = provider_copy.kwargs[a2]

        self.assertNotEqual(provider.kwargs, provider_copy.kwargs)

        self.assertIs(dependent_provider1.cls, dependent_provider_copy1.cls)
        self.assertIsNot(dependent_provider1, dependent_provider_copy1)

        self.assertIs(dependent_provider2.cls, dependent_provider_copy2.cls)
        self.assertIsNot(dependent_provider2, dependent_provider_copy2)

    def test_deepcopy_overridden(self):
        provider = providers.Dict()
        object_provider = providers.Object(object())

        provider.override(object_provider)

        provider_copy = providers.deepcopy(provider)
        object_provider_copy = provider_copy.overridden[0]

        self.assertIsNot(provider, provider_copy)
        self.assertEqual(provider.kwargs, provider_copy.kwargs)
        self.assertIsInstance(provider, providers.Dict)

        self.assertIsNot(object_provider, object_provider_copy)
        self.assertIsInstance(object_provider_copy, providers.Object)

    def test_deepcopy_with_sys_streams(self):
        provider = providers.Dict()
        provider.add_kwargs(stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)

        provider_copy = providers.deepcopy(provider)

        self.assertIsNot(provider, provider_copy)
        self.assertIsInstance(provider_copy, providers.Dict)
        self.assertIs(provider.kwargs['stdin'], sys.stdin)
        self.assertIs(provider.kwargs['stdout'], sys.stdout)
        self.assertIs(provider.kwargs['stderr'], sys.stderr)

    def test_repr(self):
        provider = providers.Dict(a1=1, a2=2)
        self.assertEqual(repr(provider),
                         '<dependency_injector.providers.'
                         'Dict({0}) at {1}>'.format(
                             repr(provider.kwargs),
                             hex(id(provider))))



"""Resources module."""

import abc
import sys
from typing import TypeVar, Generic

if sys.version_info < (3, 7):
    from typing import GenericMeta
else:
    class GenericMeta(type):
        ...


T = TypeVar('T')


class ResourceMeta(GenericMeta, abc.ABCMeta):
    def __getitem__(cls, item):
        # Spike for Python 3.6
        return cls(item)


class Resource(Generic[T], metaclass=ResourceMeta):

    @abc.abstractmethod
    def init(self, *args, **kwargs) -> T:
        ...

    @abc.abstractmethod
    def shutdown(self, resource: T) -> None:
        ...


class AsyncResource(Generic[T], metaclass=ResourceMeta):

    @abc.abstractmethod
    async def init(self, *args, **kwargs) -> T:
        ...

    @abc.abstractmethod
    async def shutdown(self, resource: T) -> None:
        ...



import contextlib
import json
import os.path
import tempfile
import unittest

import yaml
from dependency_injector import containers, providers, errors


class FromSchemaTests(unittest.TestCase):

    def test(self):
        container = containers.DynamicContainer()
        container.from_schema(
            {
                'version': '1',
                'container': {
                    'provider1': {
                        'provider': 'Factory',
                        'provides': 'list',
                        'args': [1, 2, 3],
                    },
                    'provider2': {
                        'provider': 'Factory',
                        'provides': 'dict',
                        'kwargs': {
                            'one': 'container.provider1',
                            'two': 2,
                        },
                    },
                },
            },
        )

        self.assertIsInstance(container.provider1, providers.Factory)
        self.assertIs(container.provider1.provides, list)
        self.assertEqual(container.provider1.args, (1, 2, 3))

        self.assertIsInstance(container.provider2, providers.Factory)
        self.assertIs(container.provider2.provides, dict)
        self.assertEqual(container.provider2.kwargs, {'one': container.provider1, 'two': 2})


class FromYamlSchemaTests(unittest.TestCase):

    def test(self):
        container = containers.DynamicContainer()

        with tempfile.TemporaryDirectory() as tmp_dir:
            schema_path = os.path.join(tmp_dir, 'schema.yml')
            with open(schema_path, 'w') as file:
                file.write("""
                version: "1"
                container:
                  provider1:
                    provider: Factory
                    provides: list
                    args:
                      - 1
                      - 2
                      - 3
                  provider2:
                    provider: Factory
                    provides: dict
                    kwargs:
                      one: container.provider1
                      two: 2
                """)

            container.from_yaml_schema(schema_path)

        self.assertIsInstance(container.provider1, providers.Factory)
        self.assertIs(container.provider1.provides, list)
        self.assertEqual(container.provider1.args, (1, 2, 3))

        self.assertIsInstance(container.provider2, providers.Factory)
        self.assertIs(container.provider2.provides, dict)
        self.assertEqual(container.provider2.kwargs, {'one': container.provider1, 'two': 2})

    def test_with_loader(self):
        container = containers.DynamicContainer()

        with tempfile.TemporaryDirectory() as tmp_dir:
            schema_path = os.path.join(tmp_dir, 'schema.yml')
            with open(schema_path, 'w') as file:
                file.write("""
                version: "1"
                container:
                  provider:
                    provider: Factory
                    provides: list
                    args: [1, 2, 3]
                """)

            container.from_yaml_schema(schema_path, loader=yaml.Loader)

        self.assertIsInstance(container.provider, providers.Factory)
        self.assertIs(container.provider.provides, list)
        self.assertEqual(container.provider.args, (1, 2, 3))

    def test_no_yaml_installed(self):
        @contextlib.contextmanager
        def no_yaml_module():
            containers.yaml = None
            yield
            containers.yaml = yaml

        container = containers.DynamicContainer()
        with no_yaml_module():
            with self.assertRaises(errors.Error) as error:
                container.from_yaml_schema('./no-yaml-installed.yml')

        self.assertEqual(
            error.exception.args[0],
            'Unable to load yaml schema - PyYAML is not installed. '
            'Install PyYAML or install Dependency Injector with yaml extras: '
            '"pip install dependency-injector[yaml]"',
        )


class FromJsonSchemaTests(unittest.TestCase):

    def test(self):
        container = containers.DynamicContainer()

        with tempfile.TemporaryDirectory() as tmp_dir:
            schema_path = os.path.join(tmp_dir, 'schema.json')
            with open(schema_path, 'w') as file:
                file.write(
                    json.dumps(
                        {
                            'version': '1',
                            'container': {
                                'provider1': {
                                    'provider': 'Factory',
                                    'provides': 'list',
                                    'args': [1, 2, 3],
                                },
                                'provider2': {
                                    'provider': 'Factory',
                                    'provides': 'dict',
                                    'kwargs': {
                                        'one': 'container.provider1',
                                        'two': 2,
                                    },
                                },
                            },
                        },
                        indent=4,
                    ),
                )

            container.from_json_schema(schema_path)

        self.assertIsInstance(container.provider1, providers.Factory)
        self.assertIs(container.provider1.provides, list)
        self.assertEqual(container.provider1.args, (1, 2, 3))

        self.assertIsInstance(container.provider2, providers.Factory)
        self.assertIs(container.provider2.provides, dict)
        self.assertEqual(container.provider2.kwargs, {'one': container.provider1, 'two': 2})



"""Test module for wiring."""

import sys

if 'pypy' not in sys.version.lower():
    import numpy  # noqa
    from numpy import *  # noqa

    import scipy  # noqa
    from scipy import *  # noqa

    import builtins  # noqa
    from builtins import *  # noqa



"""Dependency injector config providers unit tests."""

import contextlib
import decimal
import os
import sys
import tempfile

import unittest

from dependency_injector import containers, providers, errors
try:
    import yaml
except ImportError:
    yaml = None

try:
    import pydantic
except ImportError:
    pydantic = None


class ConfigTests(unittest.TestCase):

    def setUp(self):
        self.config = providers.Configuration(name='config')

    def tearDown(self):
        del self.config

    def test_init_optional(self):
        provider = providers.Configuration()
        provider.set_name('myconfig')
        provider.set_default({'foo': 'bar'})
        provider.set_strict(True)

        self.assertEqual(provider.get_name(), 'myconfig')
        self.assertEqual(provider.get_default(), {'foo': 'bar'})
        self.assertTrue(provider.get_strict())

    def test_set_name_returns_self(self):
        provider = providers.Configuration()
        self.assertIs(provider.set_name('myconfig'), provider)

    def test_set_default_returns_self(self):
        provider = providers.Configuration()
        self.assertIs(provider.set_default({}), provider)

    def test_set_strict_returns_self(self):
        provider = providers.Configuration()
        self.assertIs(provider.set_strict(True), provider)

    def test_default_name(self):
        config = providers.Configuration()
        self.assertEqual(config.get_name(), 'config')

    def test_providers_are_providers(self):
        self.assertTrue(providers.is_provider(self.config.a))
        self.assertTrue(providers.is_provider(self.config.a.b))
        self.assertTrue(providers.is_provider(self.config.a.b.c))
        self.assertTrue(providers.is_provider(self.config.a.b.d))

    def test_providers_are_not_delegates(self):
        self.assertFalse(providers.is_delegated(self.config.a))
        self.assertFalse(providers.is_delegated(self.config.a.b))
        self.assertFalse(providers.is_delegated(self.config.a.b.c))
        self.assertFalse(providers.is_delegated(self.config.a.b.d))

    def test_providers_identity(self):
        self.assertIs(self.config.a, self.config.a)
        self.assertIs(self.config.a.b, self.config.a.b)
        self.assertIs(self.config.a.b.c, self.config.a.b.c)
        self.assertIs(self.config.a.b.d, self.config.a.b.d)

    def test_get_name(self):
        self.assertEqual(self.config.a.b.c.get_name(), 'config.a.b.c')

    def test_providers_value_setting(self):
        a = self.config.a
        ab = self.config.a.b
        abc = self.config.a.b.c
        abd = self.config.a.b.d

        self.config.update({'a': {'b': {'c': 1, 'd': 2}}})

        self.assertEqual(a(), {'b': {'c': 1, 'd': 2}})
        self.assertEqual(ab(), {'c': 1, 'd': 2})
        self.assertEqual(abc(), 1)
        self.assertEqual(abd(), 2)

    def test_providers_with_already_set_value(self):
        self.config.update({'a': {'b': {'c': 1, 'd': 2}}})

        a = self.config.a
        ab = self.config.a.b
        abc = self.config.a.b.c
        abd = self.config.a.b.d

        self.assertEqual(a(), {'b': {'c': 1, 'd': 2}})
        self.assertEqual(ab(), {'c': 1, 'd': 2})
        self.assertEqual(abc(), 1)
        self.assertEqual(abd(), 2)

    def test_as_int(self):
        value_provider = providers.Callable(lambda value: value, self.config.test.as_int())
        self.config.from_dict({'test': '123'})

        value = value_provider()

        self.assertEqual(value, 123)

    def test_as_float(self):
        value_provider = providers.Callable(lambda value: value, self.config.test.as_float())
        self.config.from_dict({'test': '123.123'})

        value = value_provider()

        self.assertEqual(value, 123.123)

    def test_as_(self):
        value_provider = providers.Callable(
            lambda value: value,
            self.config.test.as_(decimal.Decimal),
        )
        self.config.from_dict({'test': '123.123'})

        value = value_provider()

        self.assertEqual(value, decimal.Decimal('123.123'))

    @unittest.skipIf(sys.version_info[:2] == (2, 7), 'Python 2.7 does not support this assert')
    def test_required(self):
        provider = providers.Callable(
            lambda value: value,
            self.config.a.required(),
        )
        with self.assertRaisesRegex(errors.Error, 'Undefined configuration option "config.a"'):
            provider()

    def test_required_defined_none(self):
        provider = providers.Callable(
            lambda value: value,
            self.config.a.required(),
        )
        self.config.from_dict({'a': None})
        self.assertIsNone(provider())

    def test_required_no_side_effect(self):
        _ = providers.Callable(
            lambda value: value,
            self.config.a.required(),
        )
        self.assertIsNone(self.config.a())

    def test_required_as_(self):
        provider = providers.List(
            self.config.int_test.required().as_int(),
            self.config.float_test.required().as_float(),
            self.config._as_test.required().as_(decimal.Decimal),
        )
        self.config.from_dict({'int_test': '1', 'float_test': '2.0', '_as_test': '3.0'})

        self.assertEqual(provider(), [1, 2.0, decimal.Decimal('3.0')])

    def test_providers_value_override(self):
        a = self.config.a
        ab = self.config.a.b
        abc = self.config.a.b.c
        abd = self.config.a.b.d

        self.config.override({'a': {'b': {'c': 1, 'd': 2}}})

        self.assertEqual(a(), {'b': {'c': 1, 'd': 2}})
        self.assertEqual(ab(), {'c': 1, 'd': 2})
        self.assertEqual(abc(), 1)
        self.assertEqual(abd(), 2)

    def test_configuration_option_override_and_reset_override(self):
        # Bug: https://github.com/ets-labs/python-dependency-injector/issues/319
        self.config.from_dict({'a': {'b': {'c': 1}}})

        self.assertEqual(self.config.a.b.c(), 1)

        with self.config.set('a.b.c', 'xxx'):
            self.assertEqual(self.config.a.b.c(), 'xxx')
        self.assertEqual(self.config.a.b.c(), 1)

        with self.config.a.b.c.override('yyy'):
            self.assertEqual(self.config.a.b.c(), 'yyy')

        self.assertEqual(self.config.a.b.c(), 1)

    def test_providers_with_already_overridden_value(self):
        self.config.override({'a': {'b': {'c': 1, 'd': 2}}})

        a = self.config.a
        ab = self.config.a.b
        abc = self.config.a.b.c
        abd = self.config.a.b.d

        self.assertEqual(a(), {'b': {'c': 1, 'd': 2}})
        self.assertEqual(ab(), {'c': 1, 'd': 2})
        self.assertEqual(abc(), 1)
        self.assertEqual(abd(), 2)

    def test_providers_with_default_value(self):
        self.config = providers.Configuration(
            name='config', default={'a': {'b': {'c': 1, 'd': 2}}})

        a = self.config.a
        ab = self.config.a.b
        abc = self.config.a.b.c
        abd = self.config.a.b.d

        self.assertEqual(a(), {'b': {'c': 1, 'd': 2}})
        self.assertEqual(ab(), {'c': 1, 'd': 2})
        self.assertEqual(abc(), 1)
        self.assertEqual(abd(), 2)

    def test_providers_with_default_value_overriding(self):
        self.config = providers.Configuration(
            name='config', default={'a': {'b': {'c': 1, 'd': 2}}})

        self.assertEqual(self.config.a(), {'b': {'c': 1, 'd': 2}})
        self.assertEqual(self.config.a.b(), {'c': 1, 'd': 2})
        self.assertEqual(self.config.a.b.c(), 1)
        self.assertEqual(self.config.a.b.d(), 2)

        self.config.override({'a': {'b': {'c': 3, 'd': 4}}})
        self.assertEqual(self.config.a(), {'b': {'c': 3, 'd': 4}})
        self.assertEqual(self.config.a.b(), {'c': 3, 'd': 4})
        self.assertEqual(self.config.a.b.c(), 3)
        self.assertEqual(self.config.a.b.d(), 4)

        self.config.reset_override()
        self.assertEqual(self.config.a(), {'b': {'c': 1, 'd': 2}})
        self.assertEqual(self.config.a.b(), {'c': 1, 'd': 2})
        self.assertEqual(self.config.a.b.c(), 1)
        self.assertEqual(self.config.a.b.d(), 2)

    def test_value_of_undefined_option(self):
        self.assertIsNone(self.config.a())

    @unittest.skipIf(sys.version_info[:2] == (2, 7), 'Python 2.7 does not support this assert')
    def test_value_of_undefined_option_in_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        with self.assertRaisesRegex(errors.Error, 'Undefined configuration option "config.a"'):
            self.config.a()

    @unittest.skipIf(sys.version_info[:2] == (2, 7), 'Python 2.7 does not support this assert')
    def test_value_of_undefined_option_with_root_none_in_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        self.config.override(None)
        with self.assertRaisesRegex(errors.Error, 'Undefined configuration option "config.a"'):
            self.config.a()

    def test_value_of_defined_none_option_in_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        self.config.from_dict({'a': None})
        self.assertIsNone(self.config.a())

    def test_getting_of_special_attributes(self):
        with self.assertRaises(AttributeError):
            self.config.__name__

    def test_getting_of_special_attributes_from_child(self):
        a = self.config.a
        with self.assertRaises(AttributeError):
            a.__name__

    def test_context_manager_alias(self):
        class Container(containers.DeclarativeContainer):
            config = providers.Configuration()

        container = Container()

        with container.config as cfg:
            cfg.override({'foo': 'foo', 'bar': 'bar'})

        self.assertEqual(container.config(), {'foo': 'foo', 'bar': 'bar'})
        self.assertEqual(cfg(), {'foo': 'foo', 'bar': 'bar'})
        self.assertIs(container.config, cfg)

    def test_option_context_manager_alias(self):
        class Container(containers.DeclarativeContainer):
            config = providers.Configuration()

        container = Container()

        with container.config.option as opt:
            opt.override({'foo': 'foo', 'bar': 'bar'})

        self.assertEqual(container.config(), {'option': {'foo': 'foo', 'bar': 'bar'}})
        self.assertEqual(container.config.option(), {'foo': 'foo', 'bar': 'bar'})
        self.assertEqual(opt(), {'foo': 'foo', 'bar': 'bar'})
        self.assertIs(container.config.option, opt)

    def test_missing_key(self):
        # See: https://github.com/ets-labs/python-dependency-injector/issues/358
        self.config.override(None)
        value = self.config.key()

        self.assertIsNone(value)

    def test_deepcopy(self):
        provider = providers.Configuration('config')
        provider_copy = providers.deepcopy(provider)

        self.assertIsNot(provider, provider_copy)
        self.assertIsInstance(provider, providers.Configuration)

    def test_deepcopy_from_memo(self):
        provider = providers.Configuration('config')
        provider_copy_memo = providers.Configuration('config')

        provider_copy = providers.deepcopy(
            provider, memo={id(provider): provider_copy_memo})

        self.assertIs(provider_copy, provider_copy_memo)

    def test_deepcopy_overridden(self):
        provider = providers.Configuration('config')
        object_provider = providers.Object(object())

        provider.override(object_provider)

        provider_copy = providers.deepcopy(provider)
        object_provider_copy = provider_copy.overridden[0]

        self.assertIsNot(provider, provider_copy)
        self.assertIsInstance(provider, providers.Configuration)

        self.assertIsNot(object_provider, object_provider_copy)
        self.assertIsInstance(object_provider_copy, providers.Object)

    def test_repr(self):
        self.assertEqual(repr(self.config),
                         '<dependency_injector.providers.'
                         'Configuration({0}) at {1}>'.format(
                             repr('config'),
                             hex(id(self.config))))

    def test_repr_child(self):
        self.assertEqual(repr(self.config.a.b.c),
                         '<dependency_injector.providers.'
                         'ConfigurationOption({0}) at {1}>'.format(
                             repr('config.a.b.c'),
                             hex(id(self.config.a.b.c))))


class ConfigLinkingTests(unittest.TestCase):

    class TestCore(containers.DeclarativeContainer):
        config = providers.Configuration('core')
        value_getter = providers.Callable(lambda _: _, config.value)

    class TestServices(containers.DeclarativeContainer):
        config = providers.Configuration('services')
        value_getter = providers.Callable(lambda _: _, config.value)

    def test(self):
        root_config = providers.Configuration('main')
        core = self.TestCore(config=root_config.core)
        services = self.TestServices(config=root_config.services)

        root_config.override(
            {
                'core': {
                    'value': 'core',
                },
                'services': {
                    'value': 'services',
                },
            },
        )

        self.assertEqual(core.config(), {'value': 'core'})
        self.assertEqual(core.config.value(), 'core')
        self.assertEqual(core.value_getter(), 'core')

        self.assertEqual(services.config(), {'value': 'services'})
        self.assertEqual(services.config.value(), 'services')
        self.assertEqual(services.value_getter(), 'services')

    def test_double_override(self):
        root_config = providers.Configuration('main')
        core = self.TestCore(config=root_config.core)
        services = self.TestServices(config=root_config.services)

        root_config.override(
            {
                'core': {
                    'value': 'core1',
                },
                'services': {
                    'value': 'services1',
                },
            },
        )
        root_config.override(
            {
                'core': {
                    'value': 'core2',
                },
                'services': {
                    'value': 'services2',
                },
            },
        )

        self.assertEqual(core.config(), {'value': 'core2'})
        self.assertEqual(core.config.value(), 'core2')
        self.assertEqual(core.value_getter(), 'core2')

        self.assertEqual(services.config(), {'value': 'services2'})
        self.assertEqual(services.config.value(), 'services2')
        self.assertEqual(services.value_getter(), 'services2')

    def test_reset_overriding_cache(self):
        # See: https://github.com/ets-labs/python-dependency-injector/issues/428
        class Core(containers.DeclarativeContainer):
            config = providers.Configuration()

            greetings = providers.Factory(str, config.greeting)

        class Application(containers.DeclarativeContainer):
            config = providers.Configuration()

            core = providers.Container(
                Core,
                config=config,
            )

            greetings = providers.Factory(str, config.greeting)

        container = Application()

        container.config.set('greeting', 'Hello World')
        self.assertEqual(container.greetings(), 'Hello World')
        self.assertEqual(container.core.greetings(), 'Hello World')

        container.config.set('greeting', 'Hello Bob')
        self.assertEqual(container.greetings(), 'Hello Bob')
        self.assertEqual(container.core.greetings(), 'Hello Bob')

    def test_reset_overriding_cache_for_option(self):
        # See: https://github.com/ets-labs/python-dependency-injector/issues/428
        class Core(containers.DeclarativeContainer):
            config = providers.Configuration()

            greetings = providers.Factory(str, config.greeting)

        class Application(containers.DeclarativeContainer):
            config = providers.Configuration()

            core = providers.Container(
                Core,
                config=config.option,
            )

            greetings = providers.Factory(str, config.option.greeting)

        container = Application()

        container.config.set('option.greeting', 'Hello World')
        self.assertEqual(container.greetings(), 'Hello World')
        self.assertEqual(container.core.greetings(), 'Hello World')

        container.config.set('option.greeting', 'Hello Bob')
        self.assertEqual(container.greetings(), 'Hello Bob')
        self.assertEqual(container.core.greetings(), 'Hello Bob')


class ConfigFromIniTests(unittest.TestCase):

    def setUp(self):
        self.config = providers.Configuration(name='config')

        _, self.config_file_1 = tempfile.mkstemp()
        with open(self.config_file_1, 'w') as config_file:
            config_file.write(
                '[section1]\n'
                'value1=1\n'
                '\n'
                '[section2]\n'
                'value2=2\n'
            )

        _, self.config_file_2 = tempfile.mkstemp()
        with open(self.config_file_2, 'w') as config_file:
            config_file.write(
                '[section1]\n'
                'value1=11\n'
                'value11=11\n'
                '[section3]\n'
                'value3=3\n'
            )

    def tearDown(self):
        del self.config
        os.unlink(self.config_file_1)
        os.unlink(self.config_file_2)

    def test(self):
        self.config.from_ini(self.config_file_1)

        self.assertEqual(self.config(), {'section1': {'value1': '1'}, 'section2': {'value2': '2'}})
        self.assertEqual(self.config.section1(), {'value1': '1'})
        self.assertEqual(self.config.section1.value1(), '1')
        self.assertEqual(self.config.section2(), {'value2': '2'})
        self.assertEqual(self.config.section2.value2(), '2')

    def test_option(self):
        self.config.option.from_ini(self.config_file_1)

        self.assertEqual(self.config(), {'option': {'section1': {'value1': '1'}, 'section2': {'value2': '2'}}})
        self.assertEqual(self.config.option(), {'section1': {'value1': '1'}, 'section2': {'value2': '2'}})
        self.assertEqual(self.config.option.section1(), {'value1': '1'})
        self.assertEqual(self.config.option.section1.value1(), '1')
        self.assertEqual(self.config.option.section2(), {'value2': '2'})
        self.assertEqual(self.config.option.section2.value2(), '2')

    def test_merge(self):
        self.config.from_ini(self.config_file_1)
        self.config.from_ini(self.config_file_2)

        self.assertEqual(
            self.config(),
            {
                'section1': {
                    'value1': '11',
                    'value11': '11',
                },
                'section2': {
                    'value2': '2',
                },
                'section3': {
                    'value3': '3',
                },
            },
        )
        self.assertEqual(self.config.section1(), {'value1': '11', 'value11': '11'})
        self.assertEqual(self.config.section1.value1(), '11')
        self.assertEqual(self.config.section1.value11(), '11')
        self.assertEqual(self.config.section2(), {'value2': '2'})
        self.assertEqual(self.config.section2.value2(), '2')
        self.assertEqual(self.config.section3(), {'value3': '3'})
        self.assertEqual(self.config.section3.value3(), '3')

    def test_file_does_not_exist(self):
        self.config.from_ini('./does_not_exist.ini')
        self.assertEqual(self.config(), {})

    def test_file_does_not_exist_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        with self.assertRaises(IOError):
            self.config.from_ini('./does_not_exist.ini')

    def test_option_file_does_not_exist(self):
        self.config.option.from_ini('does_not_exist.ini')
        self.assertIsNone(self.config.option.undefined())

    def test_option_file_does_not_exist_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        with self.assertRaises(IOError):
            self.config.option.from_ini('./does_not_exist.ini')

    def test_required_file_does_not_exist(self):
        with self.assertRaises(IOError):
            self.config.from_ini('./does_not_exist.ini', required=True)

    def test_required_option_file_does_not_exist(self):
        with self.assertRaises(IOError):
            self.config.option.from_ini('./does_not_exist.ini', required=True)

    def test_not_required_file_does_not_exist_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        self.config.from_ini('./does_not_exist.ini', required=False)
        self.assertEqual(self.config(), {})

    def test_not_required_option_file_does_not_exist_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        self.config.option.from_ini('./does_not_exist.ini', required=False)
        with self.assertRaises(errors.Error):
            self.config.option()


class ConfigFromIniWithEnvInterpolationTests(unittest.TestCase):

    def setUp(self):
        self.config = providers.Configuration(name='config')

        os.environ['CONFIG_TEST_ENV'] = 'test-value'
        os.environ['CONFIG_TEST_PATH'] = 'test-path'

        _, self.config_file = tempfile.mkstemp()
        with open(self.config_file, 'w') as config_file:
            config_file.write(
                '[section1]\n'
                'value1=${CONFIG_TEST_ENV}\n'
                'value2=${CONFIG_TEST_PATH}/path\n'
            )

    def tearDown(self):
        del self.config
        os.environ.pop('CONFIG_TEST_ENV', None)
        os.environ.pop('CONFIG_TEST_PATH', None)
        os.unlink(self.config_file)

    def test_env_variable_interpolation(self):
        self.config.from_ini(self.config_file)

        self.assertEqual(
            self.config(),
            {
                'section1': {
                    'value1': 'test-value',
                    'value2': 'test-path/path',
                },
            },
        )
        self.assertEqual(
            self.config.section1(),
            {
                'value1': 'test-value',
                'value2': 'test-path/path',
            },
        )
        self.assertEqual(self.config.section1.value1(), 'test-value')
        self.assertEqual(self.config.section1.value2(), 'test-path/path')

    def test_missing_envs_not_required(self):
        del os.environ['CONFIG_TEST_ENV']
        del os.environ['CONFIG_TEST_PATH']

        self.config.from_ini(self.config_file)

        self.assertEqual(
            self.config(),
            {
                'section1': {
                    'value1': '',
                    'value2': '/path',
                },
            },
        )
        self.assertEqual(
            self.config.section1(),
            {
                'value1': '',
                'value2': '/path',
            },
        )
        self.assertEqual(self.config.section1.value1(), '')
        self.assertEqual(self.config.section1.value2(), '/path')

    def test_missing_envs_required(self):
        with open(self.config_file, 'w') as config_file:
            config_file.write(
                '[section]\n'
                'undefined=${UNDEFINED}\n'
            )

        with self.assertRaises(ValueError) as context:
            self.config.from_ini(self.config_file, envs_required=True)

        self.assertEqual(
            str(context.exception),
            'Missing required environment variable "UNDEFINED"',
        )

    def test_missing_envs_strict_mode(self):
        with open(self.config_file, 'w') as config_file:
            config_file.write(
                '[section]\n'
                'undefined=${UNDEFINED}\n'
            )

        self.config.set_strict(True)
        with self.assertRaises(ValueError) as context:
            self.config.from_ini(self.config_file)

        self.assertEqual(
            str(context.exception),
            'Missing required environment variable "UNDEFINED"',
        )

    def test_option_missing_envs_not_required(self):
        del os.environ['CONFIG_TEST_ENV']
        del os.environ['CONFIG_TEST_PATH']

        self.config.option.from_ini(self.config_file)

        self.assertEqual(
            self.config.option(),
            {
                'section1': {
                    'value1': '',
                    'value2': '/path',
                },
            },
        )
        self.assertEqual(
            self.config.option.section1(),
            {
                'value1': '',
                'value2': '/path',
            },
        )
        self.assertEqual(self.config.option.section1.value1(), '')
        self.assertEqual(self.config.option.section1.value2(), '/path')

    def test_option_missing_envs_required(self):
        with open(self.config_file, 'w') as config_file:
            config_file.write(
                '[section]\n'
                'undefined=${UNDEFINED}\n'
            )

        with self.assertRaises(ValueError) as context:
            self.config.option.from_ini(self.config_file, envs_required=True)

        self.assertEqual(
            str(context.exception),
            'Missing required environment variable "UNDEFINED"',
        )

    def test_option_missing_envs_strict_mode(self):
        with open(self.config_file, 'w') as config_file:
            config_file.write(
                '[section]\n'
                'undefined=${UNDEFINED}\n'
            )

        self.config.set_strict(True)
        with self.assertRaises(ValueError) as context:
            self.config.option.from_ini(self.config_file)

        self.assertEqual(
            str(context.exception),
            'Missing required environment variable "UNDEFINED"',
        )

    def test_default_values(self):
        os.environ['DEFINED'] = 'defined'
        self.addCleanup(os.environ.pop, 'DEFINED')

        with open(self.config_file, 'w') as config_file:
            config_file.write(
                '[section]\n'
                'defined_with_default=${DEFINED:default}\n'
                'undefined_with_default=${UNDEFINED:default}\n'
                'complex=${DEFINED}/path/${DEFINED:default}/${UNDEFINED}/${UNDEFINED:default}\n'
            )

        self.config.from_ini(self.config_file)

        self.assertEqual(
            self.config.section(),
            {
                'defined_with_default': 'defined',
                'undefined_with_default': 'default',
                'complex': 'defined/path/defined//default',
            },
        )


class ConfigFromYamlTests(unittest.TestCase):

    def setUp(self):
        self.config = providers.Configuration(name='config')

        _, self.config_file_1 = tempfile.mkstemp()
        with open(self.config_file_1, 'w') as config_file:
            config_file.write(
                'section1:\n'
                '  value1: 1\n'
                '\n'
                'section2:\n'
                '  value2: 2\n'
            )

        _, self.config_file_2 = tempfile.mkstemp()
        with open(self.config_file_2, 'w') as config_file:
            config_file.write(
                'section1:\n'
                '  value1: 11\n'
                '  value11: 11\n'
                'section3:\n'
                '  value3: 3\n'
            )

    def tearDown(self):
        del self.config
        os.unlink(self.config_file_1)
        os.unlink(self.config_file_2)

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test(self):
        self.config.from_yaml(self.config_file_1)

        self.assertEqual(self.config(), {'section1': {'value1': 1}, 'section2': {'value2': 2}})
        self.assertEqual(self.config.section1(), {'value1': 1})
        self.assertEqual(self.config.section1.value1(), 1)
        self.assertEqual(self.config.section2(), {'value2': 2})
        self.assertEqual(self.config.section2.value2(), 2)

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_merge(self):
        self.config.from_yaml(self.config_file_1)
        self.config.from_yaml(self.config_file_2)

        self.assertEqual(
            self.config(),
            {
                'section1': {
                    'value1': 11,
                    'value11': 11,
                },
                'section2': {
                    'value2': 2,
                },
                'section3': {
                    'value3': 3,
                },
            },
        )
        self.assertEqual(self.config.section1(), {'value1': 11, 'value11': 11})
        self.assertEqual(self.config.section1.value1(), 11)
        self.assertEqual(self.config.section1.value11(), 11)
        self.assertEqual(self.config.section2(), {'value2': 2})
        self.assertEqual(self.config.section2.value2(), 2)
        self.assertEqual(self.config.section3(), {'value3': 3})
        self.assertEqual(self.config.section3.value3(), 3)

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_file_does_not_exist(self):
        self.config.from_yaml('./does_not_exist.yml')
        self.assertEqual(self.config(), {})

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_file_does_not_exist_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        with self.assertRaises(IOError):
            self.config.from_yaml('./does_not_exist.yml')

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_option_file_does_not_exist(self):
        self.config.option.from_yaml('./does_not_exist.yml')
        self.assertIsNone(self.config.option())

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_option_file_does_not_exist_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        with self.assertRaises(IOError):
            self.config.option.from_yaml('./does_not_exist.yml')

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_required_file_does_not_exist(self):
        with self.assertRaises(IOError):
            self.config.from_yaml('./does_not_exist.yml', required=True)

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_required_option_file_does_not_exist(self):
        with self.assertRaises(IOError):
            self.config.option.from_yaml('./does_not_exist.yml', required=True)

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_not_required_file_does_not_exist_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        self.config.from_yaml('./does_not_exist.yml', required=False)
        self.assertEqual(self.config(), {})

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_not_required_option_file_does_not_exist_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        self.config.option.from_yaml('./does_not_exist.yml', required=False)
        with self.assertRaises(errors.Error):
            self.config.option()

    def test_no_yaml_installed(self):
        @contextlib.contextmanager
        def no_yaml_module():
            yaml = providers.yaml
            providers.yaml = None

            yield

            providers.yaml = yaml

        with no_yaml_module():
            with self.assertRaises(errors.Error) as error:
                self.config.from_yaml(self.config_file_1)

        self.assertEqual(
            error.exception.args[0],
            'Unable to load yaml configuration - PyYAML is not installed. '
            'Install PyYAML or install Dependency Injector with yaml extras: '
            '"pip install dependency-injector[yaml]"',
        )

    def test_option_no_yaml_installed(self):
        @contextlib.contextmanager
        def no_yaml_module():
            yaml = providers.yaml
            providers.yaml = None

            yield

            providers.yaml = yaml

        with no_yaml_module():
            with self.assertRaises(errors.Error) as error:
                self.config.option.from_yaml(self.config_file_1)

        self.assertEqual(
            error.exception.args[0],
            'Unable to load yaml configuration - PyYAML is not installed. '
            'Install PyYAML or install Dependency Injector with yaml extras: '
            '"pip install dependency-injector[yaml]"',
        )


class ConfigFromYamlWithEnvInterpolationTests(unittest.TestCase):

    def setUp(self):
        self.config = providers.Configuration(name='config')

        os.environ['CONFIG_TEST_ENV'] = 'test-value'
        os.environ['CONFIG_TEST_PATH'] = 'test-path'

        _, self.config_file = tempfile.mkstemp()
        with open(self.config_file, 'w') as config_file:
            config_file.write(
                'section1:\n'
                '  value1: ${CONFIG_TEST_ENV}\n'
                '  value2: ${CONFIG_TEST_PATH}/path\n'
            )

    def tearDown(self):
        del self.config
        os.environ.pop('CONFIG_TEST_ENV', None)
        os.environ.pop('CONFIG_TEST_PATH', None)
        os.unlink(self.config_file)

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_env_variable_interpolation(self):
        self.config.from_yaml(self.config_file)

        self.assertEqual(
            self.config(),
            {
                'section1': {
                    'value1': 'test-value',
                    'value2': 'test-path/path',
                },
            },
        )
        self.assertEqual(
            self.config.section1(),
            {
                'value1': 'test-value',
                'value2': 'test-path/path',
            },
        )
        self.assertEqual(self.config.section1.value1(), 'test-value')
        self.assertEqual(self.config.section1.value2(), 'test-path/path')

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_missing_envs_not_required(self):
        del os.environ['CONFIG_TEST_ENV']
        del os.environ['CONFIG_TEST_PATH']

        self.config.from_yaml(self.config_file)

        self.assertEqual(
            self.config(),
            {
                'section1': {
                    'value1': None,
                    'value2': '/path',
                },
            },
        )
        self.assertEqual(
            self.config.section1(),
            {
                'value1': None,
                'value2': '/path',
            },
        )
        self.assertIsNone(self.config.section1.value1())
        self.assertEqual(self.config.section1.value2(), '/path')

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_missing_envs_required(self):
        with open(self.config_file, 'w') as config_file:
            config_file.write(
                'section:\n'
                '  undefined: ${UNDEFINED}\n'
            )

        with self.assertRaises(ValueError) as context:
            self.config.from_yaml(self.config_file, envs_required=True)

        self.assertEqual(
            str(context.exception),
            'Missing required environment variable "UNDEFINED"',
        )

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_missing_envs_strict_mode(self):
        with open(self.config_file, 'w') as config_file:
            config_file.write(
                'section:\n'
                '  undefined: ${UNDEFINED}\n'
            )

        self.config.set_strict(True)
        with self.assertRaises(ValueError) as context:
            self.config.from_yaml(self.config_file)

        self.assertEqual(
            str(context.exception),
            'Missing required environment variable "UNDEFINED"',
        )

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_option_missing_envs_not_required(self):
        del os.environ['CONFIG_TEST_ENV']
        del os.environ['CONFIG_TEST_PATH']

        self.config.option.from_yaml(self.config_file)

        self.assertEqual(
            self.config.option(),
            {
                'section1': {
                    'value1': None,
                    'value2': '/path',
                },
            },
        )
        self.assertEqual(
            self.config.option.section1(),
            {
                'value1': None,
                'value2': '/path',
            },
        )
        self.assertIsNone(self.config.option.section1.value1())
        self.assertEqual(self.config.option.section1.value2(), '/path')

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_option_missing_envs_required(self):
        with open(self.config_file, 'w') as config_file:
            config_file.write(
                'section:\n'
                '  undefined: ${UNDEFINED}\n'
            )

        with self.assertRaises(ValueError) as context:
            self.config.option.from_yaml(self.config_file, envs_required=True)

        self.assertEqual(
            str(context.exception),
            'Missing required environment variable "UNDEFINED"',
        )

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_option_missing_envs_strict_mode(self):
        with open(self.config_file, 'w') as config_file:
            config_file.write(
                'section:\n'
                '  undefined: ${UNDEFINED}\n'
            )

        self.config.set_strict(True)
        with self.assertRaises(ValueError) as context:
            self.config.option.from_yaml(self.config_file)

        self.assertEqual(
            str(context.exception),
            'Missing required environment variable "UNDEFINED"',
        )

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_default_values(self):
        os.environ['DEFINED'] = 'defined'
        self.addCleanup(os.environ.pop, 'DEFINED')

        with open(self.config_file, 'w') as config_file:
            config_file.write(
                'section:\n'
                '  defined_with_default: ${DEFINED:default}\n'
                '  undefined_with_default: ${UNDEFINED:default}\n'
                '  complex: ${DEFINED}/path/${DEFINED:default}/${UNDEFINED}/${UNDEFINED:default}\n'
            )

        self.config.from_yaml(self.config_file)

        self.assertEqual(
            self.config.section(),
            {
                'defined_with_default': 'defined',
                'undefined_with_default': 'default',
                'complex': 'defined/path/defined//default',
            },
        )

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_option_env_variable_interpolation(self):
        self.config.option.from_yaml(self.config_file)

        self.assertEqual(
            self.config.option(),
            {
                'section1': {
                    'value1': 'test-value',
                    'value2': 'test-path/path',
                },
            },
        )
        self.assertEqual(
            self.config.option.section1(),
            {
                'value1': 'test-value',
                'value2': 'test-path/path',
            },
        )
        self.assertEqual(self.config.option.section1.value1(), 'test-value')
        self.assertEqual(self.config.option.section1.value2(), 'test-path/path')

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_env_variable_interpolation_custom_loader(self):
        self.config.from_yaml(self.config_file, loader=yaml.UnsafeLoader)

        self.assertEqual(
            self.config.section1(),
            {
                'value1': 'test-value',
                'value2': 'test-path/path',
            },
        )
        self.assertEqual(self.config.section1.value1(), 'test-value')
        self.assertEqual(self.config.section1.value2(), 'test-path/path')

    @unittest.skipIf(sys.version_info[:2] == (3, 4), 'PyYAML does not support Python 3.4')
    def test_option_env_variable_interpolation_custom_loader(self):
        self.config.option.from_yaml(self.config_file, loader=yaml.UnsafeLoader)

        self.assertEqual(
            self.config.option.section1(),
            {
                'value1': 'test-value',
                'value2': 'test-path/path',
            },
        )
        self.assertEqual(self.config.option.section1.value1(), 'test-value')
        self.assertEqual(self.config.option.section1.value2(), 'test-path/path')


class ConfigFromPydanticTests(unittest.TestCase):

    def setUp(self):
        self.config = providers.Configuration(name='config')

        class Section11(pydantic.BaseModel):
            value1 = 1

        class Section12(pydantic.BaseModel):
            value2 = 2

        class Settings1(pydantic.BaseSettings):
            section1 = Section11()
            section2 = Section12()

        self.Settings1 = Settings1

        class Section21(pydantic.BaseModel):
            value1 = 11
            value11 = 11

        class Section3(pydantic.BaseModel):
            value3 = 3

        class Settings2(pydantic.BaseSettings):
            section1 = Section21()
            section3 = Section3()

        self.Settings2 = Settings2

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test(self):
        self.config.from_pydantic(self.Settings1())

        self.assertEqual(self.config(), {'section1': {'value1': 1}, 'section2': {'value2': 2}})
        self.assertEqual(self.config.section1(), {'value1': 1})
        self.assertEqual(self.config.section1.value1(), 1)
        self.assertEqual(self.config.section2(), {'value2': 2})
        self.assertEqual(self.config.section2.value2(), 2)

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_kwarg(self):
        self.config.from_pydantic(self.Settings1(), exclude={'section2'})

        self.assertEqual(self.config(), {'section1': {'value1': 1}})
        self.assertEqual(self.config.section1(), {'value1': 1})
        self.assertEqual(self.config.section1.value1(), 1)

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_merge(self):
        self.config.from_pydantic(self.Settings1())
        self.config.from_pydantic(self.Settings2())

        self.assertEqual(
            self.config(),
            {
                'section1': {
                    'value1': 11,
                    'value11': 11,
                },
                'section2': {
                    'value2': 2,
                },
                'section3': {
                    'value3': 3,
                },
            },
        )
        self.assertEqual(self.config.section1(), {'value1': 11, 'value11': 11})
        self.assertEqual(self.config.section1.value1(), 11)
        self.assertEqual(self.config.section1.value11(), 11)
        self.assertEqual(self.config.section2(), {'value2': 2})
        self.assertEqual(self.config.section2.value2(), 2)
        self.assertEqual(self.config.section3(), {'value3': 3})
        self.assertEqual(self.config.section3.value3(), 3)

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_empty_settings(self):
        self.config.from_pydantic(pydantic.BaseSettings())
        self.assertEqual(self.config(), {})

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_empty_settings_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        with self.assertRaises(ValueError):
            self.config.from_pydantic(pydantic.BaseSettings())

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_option_empty_settings(self):
        self.config.option.from_pydantic(pydantic.BaseSettings())
        self.assertEqual(self.config.option(), {})

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_option_empty_settings_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        with self.assertRaises(ValueError):
            self.config.option.from_pydantic(pydantic.BaseSettings())

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_required_empty_settings(self):
        with self.assertRaises(ValueError):
            self.config.from_pydantic(pydantic.BaseSettings(), required=True)

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_required_option_empty_settings(self):
        with self.assertRaises(ValueError):
            self.config.option.from_pydantic(pydantic.BaseSettings(), required=True)

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_not_required_empty_settings_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        self.config.from_pydantic(pydantic.BaseSettings(), required=False)
        self.assertEqual(self.config(), {})

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_not_required_option_empty_settings_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        self.config.option.from_pydantic(pydantic.BaseSettings(), required=False)
        self.assertEqual(self.config.option(), {})
        self.assertEqual(self.config(), {'option': {}})

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_not_instance_of_settings(self):
        with self.assertRaises(errors.Error) as error:
            self.config.from_pydantic({})

        self.assertEqual(
            error.exception.args[0],
            'Unable to recognize settings instance, expect "pydantic.BaseSettings", '
            'got {0} instead'.format({})
        )

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_option_not_instance_of_settings(self):
        with self.assertRaises(errors.Error) as error:
            self.config.option.from_pydantic({})

        self.assertEqual(
            error.exception.args[0],
            'Unable to recognize settings instance, expect "pydantic.BaseSettings", '
            'got {0} instead'.format({})
        )

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_subclass_instead_of_instance(self):
        with self.assertRaises(errors.Error) as error:
            self.config.from_pydantic(self.Settings1)

        self.assertEqual(
            error.exception.args[0],
            'Got settings class, but expect instance: '
            'instead "Settings1" use "Settings1()"'
        )

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_option_subclass_instead_of_instance(self):
        with self.assertRaises(errors.Error) as error:
            self.config.option.from_pydantic(self.Settings1)

        self.assertEqual(
            error.exception.args[0],
            'Got settings class, but expect instance: '
            'instead "Settings1" use "Settings1()"'
        )

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_no_pydantic_installed(self):
        @contextlib.contextmanager
        def no_pydantic_module():
            pydantic = providers.pydantic
            providers.pydantic = None

            yield

            providers.pydantic = pydantic

        with no_pydantic_module():
            with self.assertRaises(errors.Error) as error:
                self.config.from_pydantic(self.Settings1())

        self.assertEqual(
            error.exception.args[0],
            'Unable to load pydantic configuration - pydantic is not installed. '
            'Install pydantic or install Dependency Injector with pydantic extras: '
            '"pip install dependency-injector[pydantic]"',
        )

    @unittest.skipIf(sys.version_info[:2] < (3, 6), 'Pydantic supports Python 3.6+')
    def test_option_no_pydantic_installed(self):
        @contextlib.contextmanager
        def no_pydantic_module():
            pydantic = providers.pydantic
            providers.pydantic = None

            yield

            providers.pydantic = pydantic

        with no_pydantic_module():
            with self.assertRaises(errors.Error) as error:
                self.config.option.from_pydantic(self.Settings1())

        self.assertEqual(
            error.exception.args[0],
            'Unable to load pydantic configuration - pydantic is not installed. '
            'Install pydantic or install Dependency Injector with pydantic extras: '
            '"pip install dependency-injector[pydantic]"',
        )


class ConfigFromDict(unittest.TestCase):

    def setUp(self):
        self.config = providers.Configuration(name='config')

        self.config_options_1 = {
            'section1': {
                'value1': '1',
            },
            'section2': {
                'value2': '2',
            },
        }
        self.config_options_2 = {
            'section1': {
                'value1': '11',
                'value11': '11',
            },
            'section3': {
                'value3': '3',
            },
        }

    def test(self):
        self.config.from_dict(self.config_options_1)

        self.assertEqual(self.config(), {'section1': {'value1': '1'}, 'section2': {'value2': '2'}})
        self.assertEqual(self.config.section1(), {'value1': '1'})
        self.assertEqual(self.config.section1.value1(), '1')
        self.assertEqual(self.config.section2(), {'value2': '2'})
        self.assertEqual(self.config.section2.value2(), '2')

    def test_merge(self):
        self.config.from_dict(self.config_options_1)
        self.config.from_dict(self.config_options_2)

        self.assertEqual(
            self.config(),
            {
                'section1': {
                    'value1': '11',
                    'value11': '11',
                },
                'section2': {
                    'value2': '2',
                },
                'section3': {
                    'value3': '3',
                },
            },
        )
        self.assertEqual(self.config.section1(), {'value1': '11', 'value11': '11'})
        self.assertEqual(self.config.section1.value1(), '11')
        self.assertEqual(self.config.section1.value11(), '11')
        self.assertEqual(self.config.section2(), {'value2': '2'})
        self.assertEqual(self.config.section2.value2(), '2')
        self.assertEqual(self.config.section3(), {'value3': '3'})
        self.assertEqual(self.config.section3.value3(), '3')

    def test_empty_dict(self):
        self.config.from_dict({})
        self.assertEqual(self.config(), {})

    def test_option_empty_dict(self):
        self.config.option.from_dict({})
        self.assertEqual(self.config.option(), {})

    def test_empty_dict_in_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        with self.assertRaises(ValueError):
            self.config.from_dict({})

    def test_option_empty_dict_in_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        with self.assertRaises(ValueError):
            self.config.option.from_dict({})

    def test_required_empty_dict(self):
        with self.assertRaises(ValueError):
            self.config.from_dict({}, required=True)

    def test_required_option_empty_dict(self):
        with self.assertRaises(ValueError):
            self.config.option.from_dict({}, required=True)

    def test_not_required_empty_dict_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        self.config.from_dict({}, required=False)
        self.assertEqual(self.config(), {})

    def test_not_required_option_empty_dict_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        self.config.option.from_dict({}, required=False)
        self.assertEqual(self.config.option(), {})
        self.assertEqual(self.config(), {'option': {}})


class ConfigFromEnvTests(unittest.TestCase):

    def setUp(self):
        self.config = providers.Configuration(name='config')
        os.environ['CONFIG_TEST_ENV'] = 'test-value'

    def tearDown(self):
        del self.config
        del os.environ['CONFIG_TEST_ENV']

    def test(self):
        self.config.from_env('CONFIG_TEST_ENV')
        self.assertEqual(self.config(), 'test-value')

    def test_with_children(self):
        self.config.section1.value1.from_env('CONFIG_TEST_ENV')

        self.assertEqual(self.config(), {'section1': {'value1': 'test-value'}})
        self.assertEqual(self.config.section1(), {'value1': 'test-value'})
        self.assertEqual(self.config.section1.value1(), 'test-value')

    def test_default(self):
        self.config.from_env('UNDEFINED_ENV', 'default-value')
        self.assertEqual(self.config(), 'default-value')

    def test_default_none(self):
        self.config.from_env('UNDEFINED_ENV')
        self.assertIsNone(self.config())

    def test_option_default_none(self):
        self.config.option.from_env('UNDEFINED_ENV')
        self.assertIsNone(self.config.option())

    def test_undefined_in_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        with self.assertRaises(ValueError):
            self.config.from_env('UNDEFINED_ENV')

    def test_option_undefined_in_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        with self.assertRaises(ValueError):
            self.config.option.from_env('UNDEFINED_ENV')

    def test_undefined_in_strict_mode_with_default(self):
        self.config = providers.Configuration(strict=True)
        self.config.from_env('UNDEFINED_ENV', 'default-value')
        self.assertEqual(self.config(), 'default-value')

    def test_option_undefined_in_strict_mode_with_default(self):
        self.config = providers.Configuration(strict=True)
        self.config.option.from_env('UNDEFINED_ENV', 'default-value')
        self.assertEqual(self.config.option(), 'default-value')

    def test_required_undefined(self):
        with self.assertRaises(ValueError):
            self.config.from_env('UNDEFINED_ENV', required=True)

    def test_required_undefined_with_default(self):
        self.config.from_env('UNDEFINED_ENV', default='default-value', required=True)
        self.assertEqual(self.config(), 'default-value')

    def test_option_required_undefined(self):
        with self.assertRaises(ValueError):
            self.config.option.from_env('UNDEFINED_ENV', required=True)

    def test_option_required_undefined_with_default(self):
        self.config.option.from_env('UNDEFINED_ENV', default='default-value', required=True)
        self.assertEqual(self.config.option(), 'default-value')

    def test_not_required_undefined_in_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        self.config.from_env('UNDEFINED_ENV', required=False)
        self.assertIsNone(self.config())

    def test_option_not_required_undefined_in_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        self.config.option.from_env('UNDEFINED_ENV', required=False)
        self.assertIsNone(self.config.option())

    def test_not_required_undefined_with_default_in_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        self.config.from_env('UNDEFINED_ENV', default='default-value', required=False)
        self.assertEqual(self.config(), 'default-value')

    def test_option_not_required_undefined_with_default_in_strict_mode(self):
        self.config = providers.Configuration(strict=True)
        self.config.option.from_env('UNDEFINED_ENV', default='default-value', required=False)
        self.assertEqual(self.config.option(), 'default-value')


class ConfigFromValueTests(unittest.TestCase):

    def setUp(self):
        self.config = providers.Configuration(name='config')

    def test_from_value(self):
        test_value = 123321
        self.config.from_value(test_value)
        self.assertEqual(self.config(), test_value)

    def test_option_from_value(self):
        test_value_1 = 123
        test_value_2 = 321

        self.config.option1.from_value(test_value_1)
        self.config.option2.from_value(test_value_2)

        self.assertEqual(self.config(), {'option1': test_value_1, 'option2': test_value_2})
        self.assertEqual(self.config.option1(), test_value_1)
        self.assertEqual(self.config.option2(), test_value_2)



import unittest

from dependency_injector import containers, providers


class TraverseTests(unittest.TestCase):

    def test_traverse_cycled_graph(self):
        provider1 = providers.Provider()

        provider2 = providers.Provider()
        provider2.override(provider1)

        provider3 = providers.Provider()
        provider3.override(provider2)

        provider1.override(provider3)  # Cycle: provider3 -> provider2 -> provider1 -> provider3

        all_providers = list(providers.traverse(provider1))

        self.assertEqual(len(all_providers), 3)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provider3, all_providers)

    def test_traverse_types_filtering(self):
        provider1 = providers.Resource(dict)
        provider2 = providers.Resource(dict)
        provider3 = providers.Provider()

        provider = providers.Provider()

        provider.override(provider1)
        provider.override(provider2)
        provider.override(provider3)

        all_providers = list(providers.traverse(provider, types=[providers.Resource]))

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)


class ProviderTests(unittest.TestCase):

    def test_traversal_overriding(self):
        provider1 = providers.Provider()
        provider2 = providers.Provider()
        provider3 = providers.Provider()

        provider = providers.Provider()

        provider.override(provider1)
        provider.override(provider2)
        provider.override(provider3)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provider3, all_providers)

    def test_traversal_overriding_nested(self):
        provider1 = providers.Provider()

        provider2 = providers.Provider()
        provider2.override(provider1)

        provider3 = providers.Provider()
        provider3.override(provider2)

        provider = providers.Provider()
        provider.override(provider3)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provider3, all_providers)

    def test_traverse_types_filtering(self):
        provider1 = providers.Resource(dict)
        provider2 = providers.Resource(dict)
        provider3 = providers.Provider()

        provider = providers.Provider()

        provider.override(provider1)
        provider.override(provider2)
        provider.override(provider3)

        all_providers = list(provider.traverse(types=[providers.Resource]))

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)


class ObjectTests(unittest.TestCase):

    def test_traversal(self):
        provider = providers.Object('string')
        all_providers = list(provider.traverse())
        self.assertEqual(len(all_providers), 0)

    def test_traversal_provider(self):
        another_provider = providers.Provider()
        provider = providers.Object(another_provider)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 1)
        self.assertIn(another_provider, all_providers)

    def test_traversal_provider_and_overriding(self):
        another_provider_1 = providers.Provider()
        another_provider_2 = providers.Provider()
        another_provider_3 = providers.Provider()

        provider = providers.Object(another_provider_1)

        provider.override(another_provider_2)
        provider.override(another_provider_3)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(another_provider_1, all_providers)
        self.assertIn(another_provider_2, all_providers)
        self.assertIn(another_provider_3, all_providers)


class DelegateTests(unittest.TestCase):

    def test_traversal_provider(self):
        another_provider = providers.Provider()
        provider = providers.Delegate(another_provider)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 1)
        self.assertIn(another_provider, all_providers)

    def test_traversal_provider_and_overriding(self):
        provider1 = providers.Provider()
        provider2 = providers.Provider()

        provider3 = providers.Provider()
        provider3.override(provider2)

        provider = providers.Delegate(provider1)

        provider.override(provider3)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provider3, all_providers)


class DependencyTests(unittest.TestCase):

    def test_traversal(self):
        provider = providers.Dependency()
        all_providers = list(provider.traverse())
        self.assertEqual(len(all_providers), 0)

    def test_traversal_default(self):
        another_provider = providers.Provider()
        provider = providers.Dependency(default=another_provider)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 1)
        self.assertIn(another_provider, all_providers)

    def test_traversal_overriding(self):
        provider1 = providers.Provider()

        provider2 = providers.Provider()
        provider2.override(provider1)

        provider = providers.Dependency()
        provider.override(provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)


class DependenciesContainerTests(unittest.TestCase):

    def test_traversal(self):
        provider = providers.DependenciesContainer()
        all_providers = list(provider.traverse())
        self.assertEqual(len(all_providers), 0)

    def test_traversal_default(self):
        another_provider = providers.Provider()
        provider = providers.DependenciesContainer(default=another_provider)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 1)
        self.assertIn(another_provider, all_providers)

    def test_traversal_fluent_interface(self):
        provider = providers.DependenciesContainer()
        provider1 = provider.provider1
        provider2 = provider.provider2

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traversal_overriding(self):
        provider1 = providers.Provider()
        provider2 = providers.Provider()
        provider3 = providers.DependenciesContainer(
            provider1=provider1,
            provider2=provider2,
        )

        provider = providers.DependenciesContainer()
        provider.override(provider3)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 5)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provider3, all_providers)
        self.assertIn(provider.provider1, all_providers)
        self.assertIn(provider.provider2, all_providers)


class CallableTests(unittest.TestCase):

    def test_traverse(self):
        provider = providers.Callable(dict)
        all_providers = list(provider.traverse())
        self.assertEqual(len(all_providers), 0)

    def test_traverse_args(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')
        provider = providers.Callable(list, 'foo', provider1, provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_kwargs(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')
        provider = providers.Callable(dict, foo='foo', bar=provider1, baz=provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_overridden(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')

        provider = providers.Callable(dict, 'foo')
        provider.override(provider1)
        provider.override(provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_provides(self):
        provider1 = providers.Callable(list)
        provider2 = providers.Object('bar')
        provider3 = providers.Object('baz')

        provider = providers.Callable(provider1, provider2)
        provider.override(provider3)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provider3, all_providers)


class ConfigurationTests(unittest.TestCase):

    def test_traverse(self):
        config = providers.Configuration(default={'option1': {'option2': 'option2'}})
        option1 = config.option1
        option2 = config.option1.option2
        option3 = config.option1[config.option1.option2]

        all_providers = list(config.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(option1, all_providers)
        self.assertIn(option2, all_providers)
        self.assertIn(option3, all_providers)

    def test_traverse_typed(self):
        config = providers.Configuration()
        option = config.option
        typed_option = config.option.as_int()

        all_providers = list(typed_option.traverse())

        self.assertEqual(len(all_providers), 1)
        self.assertIn(option, all_providers)

    def test_traverse_overridden(self):
        options = {'option1': {'option2': 'option2'}}
        config = providers.Configuration()
        config.from_dict(options)

        all_providers = list(config.traverse())

        self.assertEqual(len(all_providers), 1)
        overridden, = all_providers
        self.assertEqual(overridden(), options)
        self.assertIs(overridden, config.last_overriding)

    def test_traverse_overridden_option_1(self):
        options = {'option2': 'option2'}
        config = providers.Configuration()
        config.option1.from_dict(options)

        all_providers = list(config.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(config.option1, all_providers)
        self.assertIn(config.last_overriding, all_providers)

    def test_traverse_overridden_option_2(self):
        options = {'option2': 'option2'}
        config = providers.Configuration()
        config.option1.from_dict(options)

        all_providers = list(config.option1.traverse())

        self.assertEqual(len(all_providers), 0)


class FactoryTests(unittest.TestCase):

    def test_traverse(self):
        provider = providers.Factory(dict)
        all_providers = list(provider.traverse())
        self.assertEqual(len(all_providers), 0)

    def test_traverse_args(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')
        provider = providers.Factory(list, 'foo', provider1, provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_kwargs(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')
        provider = providers.Factory(dict, foo='foo', bar=provider1, baz=provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_attributes(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')
        provider = providers.Factory(dict)
        provider.add_attributes(foo='foo', bar=provider1, baz=provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_overridden(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')

        provider = providers.Factory(dict, 'foo')
        provider.override(provider1)
        provider.override(provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_provides(self):
        provider1 = providers.Callable(list)
        provider2 = providers.Object('bar')
        provider3 = providers.Object('baz')

        provider = providers.Factory(provider1, provider2)
        provider.override(provider3)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provider3, all_providers)


class FactoryAggregateTests(unittest.TestCase):

    def test_traverse(self):
        factory1 = providers.Factory(dict)
        factory2 = providers.Factory(list)
        provider = providers.FactoryAggregate(factory1=factory1, factory2=factory2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(factory1, all_providers)
        self.assertIn(factory2, all_providers)


class BaseSingletonTests(unittest.TestCase):

    def test_traverse(self):
        provider = providers.Singleton(dict)
        all_providers = list(provider.traverse())
        self.assertEqual(len(all_providers), 0)

    def test_traverse_args(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')
        provider = providers.Singleton(list, 'foo', provider1, provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_kwargs(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')
        provider = providers.Singleton(dict, foo='foo', bar=provider1, baz=provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_attributes(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')
        provider = providers.Singleton(dict)
        provider.add_attributes(foo='foo', bar=provider1, baz=provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_overridden(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')

        provider = providers.Singleton(dict, 'foo')
        provider.override(provider1)
        provider.override(provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_provides(self):
        provider1 = providers.Callable(list)
        provider2 = providers.Object('bar')
        provider3 = providers.Object('baz')

        provider = providers.Singleton(provider1, provider2)
        provider.override(provider3)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provider3, all_providers)


class ListTests(unittest.TestCase):

    def test_traverse_args(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')
        provider = providers.List('foo', provider1, provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_overridden(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')
        provider3 = providers.List(provider1, provider2)

        provider = providers.List('foo')
        provider.override(provider3)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provider3, all_providers)


class DictTests(unittest.TestCase):

    def test_traverse_kwargs(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')
        provider = providers.Dict(foo='foo', bar=provider1, baz=provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_overridden(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')
        provider3 = providers.Dict(bar=provider1, baz=provider2)

        provider = providers.Dict(foo='foo')
        provider.override(provider3)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provider3, all_providers)


class ResourceTests(unittest.TestCase):

    def test_traverse(self):
        provider = providers.Resource(dict)
        all_providers = list(provider.traverse())
        self.assertEqual(len(all_providers), 0)

    def test_traverse_args(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')
        provider = providers.Resource(list, 'foo', provider1, provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_kwargs(self):
        provider1 = providers.Object('bar')
        provider2 = providers.Object('baz')
        provider = providers.Resource(dict, foo='foo', bar=provider1, baz=provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_overridden(self):
        provider1 = providers.Resource(list)
        provider2 = providers.Resource(tuple)

        provider = providers.Resource(dict, 'foo')
        provider.override(provider1)
        provider.override(provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_provides(self):
        provider1 = providers.Callable(list)

        provider = providers.Resource(provider1)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 1)
        self.assertIn(provider1, all_providers)


class ContainerTests(unittest.TestCase):

    def test_traverse(self):
        class Container(containers.DeclarativeContainer):
            provider1 = providers.Callable(list)
            provider2 = providers.Callable(dict)

        provider = providers.Container(Container)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertEqual(
            {provider.provides for provider in all_providers},
            {list,  dict},
        )

    def test_traverse_overridden(self):
        class Container1(containers.DeclarativeContainer):
            provider1 = providers.Callable(list)
            provider2 = providers.Callable(dict)

        class Container2(containers.DeclarativeContainer):
            provider1 = providers.Callable(tuple)
            provider2 = providers.Callable(str)

        container2 = Container2()

        provider = providers.Container(Container1)
        provider.override(container2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 5)
        self.assertEqual(
            {
                provider.provides
                for provider in all_providers
                if isinstance(provider, providers.Callable)
            },
            {list, dict, tuple, str},
        )
        self.assertIn(provider.last_overriding, all_providers)
        self.assertIs(provider.last_overriding(), container2)


class SelectorTests(unittest.TestCase):

    def test_traverse(self):
        switch = lambda: 'provider1'
        provider1 = providers.Callable(list)
        provider2 = providers.Callable(dict)

        provider = providers.Selector(
            switch,
            provider1=provider1,
            provider2=provider2,
        )

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_switch(self):
        switch = providers.Callable(lambda: 'provider1')
        provider1 = providers.Callable(list)
        provider2 = providers.Callable(dict)

        provider = providers.Selector(
            switch,
            provider1=provider1,
            provider2=provider2,
        )

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(switch, all_providers)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)

    def test_traverse_overridden(self):
        provider1 = providers.Callable(list)
        provider2 = providers.Callable(dict)
        selector1 = providers.Selector(lambda: 'provider1', provider1=provider1)

        provider = providers.Selector(
            lambda: 'provider2',
            provider2=provider2,
        )
        provider.override(selector1)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(selector1, all_providers)


class ProvidedInstanceTests(unittest.TestCase):

    def test_traverse(self):
        provider1 = providers.Provider()
        provider = provider1.provided

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 1)
        self.assertIn(provider1, all_providers)

    def test_traverse_overridden(self):
        provider1 = providers.Provider()
        provider2 = providers.Provider()

        provider = provider1.provided
        provider.override(provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)


class AttributeGetterTests(unittest.TestCase):

    def test_traverse(self):
        provider1 = providers.Provider()
        provided = provider1.provided
        provider = provided.attr

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provided, all_providers)

    def test_traverse_overridden(self):
        provider1 = providers.Provider()
        provided = provider1.provided
        provider2 = providers.Provider()

        provider = provided.attr
        provider.override(provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provided, all_providers)


class ItemGetterTests(unittest.TestCase):

    def test_traverse(self):
        provider1 = providers.Provider()
        provided = provider1.provided
        provider = provided['item']

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 2)
        self.assertIn(provider1, all_providers)
        self.assertIn(provided, all_providers)

    def test_traverse_overridden(self):
        provider1 = providers.Provider()
        provided = provider1.provided
        provider2 = providers.Provider()

        provider = provided['item']
        provider.override(provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provided, all_providers)


class MethodCallerTests(unittest.TestCase):

    def test_traverse(self):
        provider1 = providers.Provider()
        provided = provider1.provided
        method = provided.method
        provider = method.call()

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 3)
        self.assertIn(provider1, all_providers)
        self.assertIn(provided, all_providers)
        self.assertIn(method, all_providers)

    def test_traverse_args(self):
        provider1 = providers.Provider()
        provided = provider1.provided
        method = provided.method
        provider2 = providers.Provider()
        provider = method.call('foo', provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 4)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provided, all_providers)
        self.assertIn(method, all_providers)

    def test_traverse_kwargs(self):
        provider1 = providers.Provider()
        provided = provider1.provided
        method = provided.method
        provider2 = providers.Provider()
        provider = method.call(foo='foo', bar=provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 4)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provided, all_providers)
        self.assertIn(method, all_providers)

    def test_traverse_overridden(self):
        provider1 = providers.Provider()
        provided = provider1.provided
        method = provided.method
        provider2 = providers.Provider()

        provider = method.call()
        provider.override(provider2)

        all_providers = list(provider.traverse())

        self.assertEqual(len(all_providers), 4)
        self.assertIn(provider1, all_providers)
        self.assertIn(provider2, all_providers)
        self.assertIn(provided, all_providers)
        self.assertIn(method, all_providers)



"""Test utils."""

import asyncio
import contextlib
import sys
import gc
import unittest


def run(main):
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(main)


def setup_test_loop(
        loop_factory=asyncio.new_event_loop
) -> asyncio.AbstractEventLoop:
    loop = loop_factory()
    try:
        module = loop.__class__.__module__
        skip_watcher = 'uvloop' in module
    except AttributeError:  # pragma: no cover
        # Just in case
        skip_watcher = True
    asyncio.set_event_loop(loop)
    if sys.platform != 'win32' and not skip_watcher:
        policy = asyncio.get_event_loop_policy()
        watcher = asyncio.SafeChildWatcher()  # type: ignore
        watcher.attach_loop(loop)
        with contextlib.suppress(NotImplementedError):
            policy.set_child_watcher(watcher)
    return loop


def teardown_test_loop(loop: asyncio.AbstractEventLoop, fast: bool = False) -> None:
    closed = loop.is_closed()
    if not closed:
        loop.call_soon(loop.stop)
        loop.run_forever()
        loop.close()

    if not fast:
        gc.collect()

    asyncio.set_event_loop(None)


class AsyncTestCase(unittest.TestCase):

    def setUp(self):
        self.loop = setup_test_loop()

    def tearDown(self):
        teardown_test_loop(self.loop)

    def _run(self, f):
        return self.loop.run_until_complete(f)



"""Dependency injector provided instance provider unit tests."""

import unittest

from dependency_injector import containers, providers


class Service:
    def __init__(self, value):
        self.value = value
        self.values = [self.value]

    def __call__(self):
        return self.value

    def __getitem__(self, item):
        return self.values[item]

    def get_value(self):
        return self.value

    def get_closure(self):
        def closure():
            return self.value
        return closure


class Client:
    def __init__(self, value):
        self.value = value


class Container(containers.DeclarativeContainer):

    service = providers.Singleton(Service, value='foo')

    client_attribute = providers.Factory(
        Client,
        value=service.provided.value,
    )

    client_item = providers.Factory(
        Client,
        value=service.provided[0],
    )

    client_attribute_item = providers.Factory(
        Client,
        value=service.provided.values[0],
    )

    client_method_call = providers.Factory(
        Client,
        value=service.provided.get_value.call(),
    )
    client_method_closure_call = providers.Factory(
        Client,
        value=service.provided.get_closure.call().call(),
    )

    client_provided_call = providers.Factory(
        Client,
        value=service.provided.call(),
    )


class ProvidedInstanceTests(unittest.TestCase):

    def setUp(self):
        self.container = Container()

    def test_is_provider(self):
        self.assertTrue(providers.is_provider(self.container.service.provided))

    def test_attribute(self):
        client = self.container.client_attribute()
        self.assertEqual(client.value, 'foo')

    def test_item(self):
        client = self.container.client_item()
        self.assertEqual(client.value, 'foo')

    def test_attribute_item(self):
        client = self.container.client_attribute_item()
        self.assertEqual(client.value, 'foo')

    def test_method_call(self):
        client = self.container.client_method_call()
        self.assertEqual(client.value, 'foo')

    def test_method_closure_call(self):
        client = self.container.client_method_closure_call()
        self.assertEqual(client.value, 'foo')

    def test_provided_call(self):
        client = self.container.client_provided_call()
        self.assertEqual(client.value, 'foo')

    def test_call_overridden(self):
        value = 'bar'
        with self.container.service.override(Service(value)):
            self.assertEqual(self.container.client_attribute().value, value)
            self.assertEqual(self.container.client_item().value, value)
            self.assertEqual(self.container.client_attribute_item().value, value)
            self.assertEqual(self.container.client_method_call().value, value)

    def test_repr_provided_instance(self):
        provider = self.container.service.provided
        self.assertEqual(
            'ProvidedInstance(\'{0}\')'.format(repr(self.container.service)),
            repr(provider),
        )

    def test_repr_attribute_getter(self):
        provider = self.container.service.provided.value
        self.assertEqual(
            'AttributeGetter(\'value\')',
            repr(provider),
        )

    def test_repr_item_getter(self):
        provider = self.container.service.provided['test-test']
        self.assertEqual(
            'ItemGetter(\'test-test\')',
            repr(provider),
        )


class LazyInitTests(unittest.TestCase):

    def test_provided_instance(self):
        provides = providers.Object(object())
        provider = providers.ProvidedInstance()
        provider.set_provides(provides)
        self.assertIs(provider.provides, provides)
        self.assertIs(provider.set_provides(providers.Provider()), provider)

    def test_attribute_getter(self):
        provides = providers.Object(object())
        provider = providers.AttributeGetter()
        provider.set_provides(provides)
        provider.set_name('__dict__')
        self.assertIs(provider.provides, provides)
        self.assertEqual(provider.name, '__dict__')
        self.assertIs(provider.set_provides(providers.Provider()), provider)
        self.assertIs(provider.set_name('__dict__'), provider)

    def test_item_getter(self):
        provides = providers.Object({'foo': 'bar'})
        provider = providers.ItemGetter()
        provider.set_provides(provides)
        provider.set_name('foo')
        self.assertIs(provider.provides, provides)
        self.assertEqual(provider.name, 'foo')
        self.assertIs(provider.set_provides(providers.Provider()), provider)
        self.assertIs(provider.set_name('foo'), provider)

    def test_method_caller(self):
        provides = providers.Object(lambda: 42)
        provider = providers.MethodCaller()
        provider.set_provides(provides)
        self.assertIs(provider.provides, provides)
        self.assertEqual(provider(), 42)
        self.assertIs(provider.set_provides(providers.Provider()), provider)


class ProvidedInstancePuzzleTests(unittest.TestCase):

    def test_puzzled(self):
        service = providers.Singleton(Service, value='foo-bar')

        dependency = providers.Object(
            {
                'a': {
                    'b': {
                        'c1': 10,
                        'c2': lambda arg: {'arg': arg}
                    },
                },
            },
        )

        test_list = providers.List(
            dependency.provided['a']['b']['c1'],
            dependency.provided['a']['b']['c2'].call(22)['arg'],
            dependency.provided['a']['b']['c2'].call(service)['arg'],
            dependency.provided['a']['b']['c2'].call(service)['arg'].value,
            dependency.provided['a']['b']['c2'].call(service)['arg'].get_value.call(),
        )

        result = test_list()

        self.assertEqual(
            result,
            [
                10,
                22,
                service(),
                'foo-bar',
                'foo-bar',
            ],
        )


class ProvidedInstanceInBaseClassTests(unittest.TestCase):

    def test_provided_attribute(self):
        provider = providers.Provider()
        assert isinstance(provider.provided, providers.ProvidedInstance)



import unittest

from dependency_injector import providers

from .singleton_common import Example, _BaseSingletonTestCase


class ContextLocalSingletonTests(_BaseSingletonTestCase, unittest.TestCase):

    singleton_cls = providers.ContextLocalSingleton

    def test_repr(self):
        provider = providers.ContextLocalSingleton(Example)

        self.assertEqual(repr(provider),
                         '<dependency_injector.providers.'
                         'ContextLocalSingleton({0}) at {1}>'.format(
                             repr(Example),
                             hex(id(provider))))

    def test_reset(self):
        provider = providers.ContextLocalSingleton(Example)

        instance1 = provider()
        self.assertIsInstance(instance1, Example)

        provider.reset()

        instance2 = provider()
        self.assertIsInstance(instance2, Example)

        self.assertIsNot(instance1, instance2)

    def test_reset_clean(self):
        provider = providers.ContextLocalSingleton(Example)
        instance1 = provider()

        provider.reset()
        provider.reset()

        instance2 = provider()
        self.assertIsNot(instance1, instance2)



import unittest

# Runtime import to avoid syntax errors in samples on Python < 3.5 and reach top-dir
import os
_TOP_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        '../',
    )),
)
_SAMPLES_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        '../samples/',
    )),
)
import sys
sys.path.append(_TOP_DIR)
sys.path.append(_SAMPLES_DIR)

from wiringflask import web


class WiringFlaskTest(unittest.TestCase):

    def test(self):
        client = web.app.test_client()

        with web.app.app_context():
            response = client.get('/')

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, b'{"result":"Ok"}\n')



from httpx import AsyncClient

# Runtime import to avoid syntax errors in samples on Python < 3.5 and reach top-dir
import os
_TOP_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        '../',
    )),
)
_SAMPLES_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        '../samples/',
    )),
)
import sys
sys.path.append(_TOP_DIR)
sys.path.append(_SAMPLES_DIR)

from asyncutils import AsyncTestCase

from wiringfastapi import web


class WiringFastAPITest(AsyncTestCase):

    client: AsyncClient

    def setUp(self) -> None:
        super().setUp()
        self.client = AsyncClient(app=web.app, base_url='http://test')

    def tearDown(self) -> None:
        self._run(self.client.aclose())
        super().tearDown()

    def test_depends_marker_injection(self):
        class ServiceMock:
            async def process(self):
                return 'Foo'

        with web.container.service.override(ServiceMock()):
            response = self._run(self.client.get('/'))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {'result': 'Foo'})

    def test_depends_injection(self):
        response = self._run(self.client.get('/auth', auth=('john_smith', 'secret')))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {'username': 'john_smith', 'password': 'secret'})



import unittest

from dependency_injector import containers


class SomeClass:
    ...


class TypesTest(unittest.TestCase):

    def test_declarative(self):
        container: containers.Container = containers.DeclarativeContainer()
        self.assertIsInstance(container, containers.Container)

    def test_dynamic(self):
        container: containers.Container = containers.DynamicContainer()
        self.assertIsInstance(container, containers.Container)



"""Dependency injector provider utils unit tests."""

import unittest

from dependency_injector import (
    providers,
    errors,
)


class IsProviderTests(unittest.TestCase):

    def test_with_instance(self):
        self.assertTrue(providers.is_provider(providers.Provider()))

    def test_with_class(self):
        self.assertFalse(providers.is_provider(providers.Provider))

    def test_with_string(self):
        self.assertFalse(providers.is_provider('some_string'))

    def test_with_object(self):
        self.assertFalse(providers.is_provider(object()))

    def test_with_subclass_instance(self):
        class SomeProvider(providers.Provider):
            pass

        self.assertTrue(providers.is_provider(SomeProvider()))

    def test_with_class_with_getattr(self):
        class SomeClass(object):
            def __getattr__(self, _):
                return False

        self.assertFalse(providers.is_provider(SomeClass()))


class EnsureIsProviderTests(unittest.TestCase):

    def test_with_instance(self):
        provider = providers.Provider()
        self.assertIs(providers.ensure_is_provider(provider), provider)

    def test_with_class(self):
        self.assertRaises(errors.Error,
                          providers.ensure_is_provider,
                          providers.Provider)

    def test_with_string(self):
        self.assertRaises(errors.Error,
                          providers.ensure_is_provider,
                          'some_string')

    def test_with_object(self):
        self.assertRaises(errors.Error, providers.ensure_is_provider, object())
